(* *********************************************************************)
(*                                                                     *)
(*              The Compcert verified compiler                         *)
(*                                                                     *)
(*          Xavier Leroy, INRIA Paris-Rocquencourt                     *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique.  All rights reserved.  This file is distributed       *)
(*  under the terms of the INRIA Non-Commercial License Agreement.     *)
(*                                                                     *)
(* *********************************************************************)

(** Correctness proof for Cminor generation. *)

Require Import Coq.Program.Equality.
Require Import FSets.
Require Import Permutation.
Require Import Coqlib.
Require Intv.
Require Import Errors.
Require Import Maps.
Require Import AST.
Require Import Integers.
Require Import Floats.
Require Import Values.
Require Import Memory.
Require Import Events.
Require Import Globalenvs.
Require Import Smallstep.
Require Import Switch.
Require Import Csharpminor.
Require Import Cminor.
Require Import Cminorgen.
Require Import StrongInject.

Open Local Scope error_monad_scope.

Section TRANSLATION.

Context `{M: Mem.MEM}.

Variable prog: Csharpminor.program.
Variable valid_region_alloc': ident -> list (ident * (Z * Z * Z)) -> Prop.
Let ge : Csharpminor.genv := Genv.globalenv prog.

(*
Hypothesis function_symbols:
  forall b f,
    Genv.find_funct_ptr ge b = Some (Internal f) ->
    forall s,
      Genv.find_symbol ge s = Some b ->
      f.(Csharpminor.fn_id) = s.
*)

Inductive valid_region_alloc (fi: ident) (l: list (ident * (Z * Z * Z))) : Prop :=
| valid_region_alloc_some
    b f
    (HFUN: Genv.find_funct_ptr ge b = Some (Internal f))
    (HID: f.(Csharpminor.fn_id) = fi)
    (HOFF: check_offsets nil f.(Csharpminor.fn_vars) = OK l)
| valid_region_alloc_none
    (HFUN: forall b f, Genv.find_funct_ptr ge b = Some (Internal f) -> f.(Csharpminor.fn_id) <> fi)
    (HVALID: valid_region_alloc' fi l)
.

Context `{SI: Mem.StrongInject _ (M := M) (valid_region_alloc := valid_region_alloc)}.

Variable tprog: program.

Hypothesis TRANSL: transl_program prog = OK tprog.
Let tge: genv := Genv.globalenv tprog.

Lemma symbols_preserved:
  forall (s: ident), Genv.find_symbol tge s = Genv.find_symbol ge s.
Proof (Genv.find_symbol_transf_partial transl_fundef _ TRANSL).

Lemma function_ptr_translated:
  forall (b: block) (f: Csharpminor.fundef),
  Genv.find_funct_ptr ge b = Some f ->
  exists tf,
  Genv.find_funct_ptr tge b = Some tf /\ transl_fundef f = OK tf.
Proof (Genv.find_funct_ptr_transf_partial transl_fundef _ TRANSL).

Lemma functions_translated:
  forall (v: val) (f: Csharpminor.fundef),
  Genv.find_funct ge v = Some f ->
  exists tf,
  Genv.find_funct tge v = Some tf /\ transl_fundef f = OK tf.
Proof (Genv.find_funct_transf_partial transl_fundef _ TRANSL).

Lemma varinfo_preserved:
  forall b, Genv.find_var_info tge b = Genv.find_var_info ge b.
Proof (Genv.find_var_info_transf_partial transl_fundef _ TRANSL).

Lemma sig_preserved:
  forall f tf,
  transl_fundef f = OK tf -> 
  Cminor.funsig tf = Csharpminor.funsig f.
Proof.
  intros until tf; destruct f; simpl in *.
  unfold transl_function, transl_function_aux.
  destruct (zle (stack_size (Csharpminor.fn_vars f)) Int.max_unsigned);  try discriminate.
  destruct (check_offsets nil (Csharpminor.fn_vars f)); try discriminate; simpl.
  destruct (transl_stmt (assign_variables_aux (PTree.empty Z) l0) nil (Csharpminor.fn_body f)); try discriminate.
  simpl. injection 1; intros; subst; reflexivity.
  injection 1; intros; subst; reflexivity.
Qed.

(** * Derived properties of memory operations *)

Lemma nextblock_storev:
  forall chunk m addr v m',
  Mem.storev chunk m addr v = Some m' -> Mem.nextblock m' = Mem.nextblock m.
Proof.
  unfold Mem.storev; intros. destruct addr; try discriminate. 
  eapply Mem.nextblock_store; eauto.
Qed.
  

(** * Correspondence between C#minor's and Cminor's environments and memory states *)

(** In C#minor, every variable is stored in a separate memory block.
  In the corresponding Cminor code, these variables become sub-blocks
  of the stack data block.  We capture these changes in memory via a
  memory injection [f]: 
  [f b = Some(b', ofs)] means that C#minor block [b] corresponds
  to a sub-block of Cminor block [b] at offset [ofs].

  A memory injection [f] defines a relation [val_inject f] between
  values and a relation [Mem.inject f] between memory states.  These
  relations will be used intensively in our proof of simulation
  between C#minor and Cminor executions. *)

(** ** Matching between Cshaprminor's temporaries and Cminor's variables *)

Definition match_temps (f: meminj) (le: Csharpminor.temp_env) (te: env) : Prop :=
  forall id v, le!id = Some v -> exists v', te!(id) = Some v' /\ val_inject f v v'.

Lemma match_temps_invariant:
  forall f f' le te,
  match_temps f le te ->
  inject_incr f f' ->
  match_temps f' le te.
Proof.
  intros; red; intros. destruct (H _ _ H1) as [v' [A B]]. exists v'; eauto.
Qed.

Lemma match_temps_assign:
  forall f le te id v tv,
  match_temps f le te ->
  val_inject f v tv ->
  match_temps f (PTree.set id v le) (PTree.set id tv te).
Proof.
  intros; red; intros. rewrite PTree.gsspec in *. destruct (peq id0 id).
  inv H1. exists tv; auto.
  eauto.
Qed.

(** ** Matching between C#minor's variable environment and Cminor's stack pointer *)

Inductive match_var (f: meminj) (sp: block): option (block * Z) -> option Z -> Prop :=
  | match_var_local: forall b sz ofs,
      val_inject f (Vptr b Int.zero) (Vptr sp (Int.repr ofs)) ->
      match_var f sp (Some(b, sz)) (Some ofs)
  | match_var_global:
      match_var f sp None None.

(** Matching between a C#minor environment [e] and a Cminor
  stack pointer [sp]. The [lo] and [hi] parameters delimit the range
  of addresses for the blocks referenced from [te]. *)

Record match_env (f: meminj) (cenv: compilenv)
                 (e: Csharpminor.env) (sp: block)
                 (lo hi: Z) : Prop :=
  mk_match_env {

(** C#minor local variables match sub-blocks of the Cminor stack data block. *)
    me_vars:
      forall id, match_var f sp (e!id) (cenv!id);

(** [lo, hi] is a proper interval. *)
    me_low_high:
      lo <= hi;

(** Every block appearing in the C#minor environment [e] must be
  in the range [lo, hi]. *)
    me_bounded:
      forall id b sz, PTree.get id e = Some(b, sz) -> lo <= b < hi;

(** All blocks mapped to sub-blocks of the Cminor stack data must be
    images of variables from the C#minor environment [e] *)
    me_inv:
      forall b delta,
      f b = Some(sp, delta) ->
      exists id, exists sz, PTree.get id e = Some(b, sz);

(** All C#minor blocks below [lo] (i.e. allocated before the blocks
  referenced from [e]) must map to blocks that are below [sp]
  (i.e. allocated before the stack data for the current Cminor function). *)
    me_incr:
      forall b tb delta,
      f b = Some(tb, delta) -> b < lo -> tb < sp
    ;

    me_sp:
      sp = lo - 1
  }.

Ltac geninv x := 
  let H := fresh in (generalize x; intro H; inv H).

Lemma match_env_invariant:
  forall f1 cenv e sp lo hi f2,
  match_env f1 cenv e sp lo hi ->
  inject_incr f1 f2 ->
  (forall b delta, f2 b = Some(sp, delta) -> f1 b = Some(sp, delta)) ->
  (forall b, b < lo -> f2 b = f1 b) ->
  match_env f2 cenv e sp lo hi.
Proof.
  intros. destruct H. constructor; auto.
(* vars *)
  intros. geninv (me_vars0 id); econstructor; eauto.
(* bounded *)
  intros. eauto. 
(* below *)
  intros. rewrite H2 in H; eauto.
Qed.

(** [match_env] and external calls *)

Remark inject_incr_separated_same:
  forall f1 f2 m1 m1',
  inject_incr f1 f2 -> inject_separated f1 f2 m1 m1' ->
  forall b, Mem.valid_block m1 b -> f2 b = f1 b.
Proof.
  intros. case_eq (f1 b).
  intros [b' delta] EQ. apply H; auto. 
  intros EQ. case_eq (f2 b).
  intros [b'1 delta1] EQ1. exploit H0; eauto. intros [C D]. contradiction.
  auto.
Qed.

Remark inject_incr_separated_same':
  forall f1 f2 m1 m1',
  inject_incr f1 f2 -> inject_separated f1 f2 m1 m1' ->
  forall b b' delta,
  f2 b = Some(b', delta) -> Mem.valid_block m1' b' -> f1 b = Some(b', delta).
Proof.
  intros. case_eq (f1 b).
  intros [b'1 delta1] EQ. exploit H; eauto. congruence.
  intros. exploit H0; eauto. intros [C D]. contradiction.
Qed.

(** The sizes of blocks appearing in [e] are respected. *)

Definition match_bounds (e: Csharpminor.env) (m: mem) : Prop :=
  forall id b sz ofs p,
  PTree.get id e = Some(b, sz) -> Mem.perm m b ofs Max p -> 0 <= ofs < sz.

Lemma match_bounds_invariant:
  forall e m1 m2,
  match_bounds e m1 ->
  (forall id b sz ofs p,
   PTree.get id e = Some(b, sz) -> Mem.perm m2 b ofs Max p -> Mem.perm m1 b ofs Max p) ->
  match_bounds e m2.
Proof.
  intros; red; intros. eapply H; eauto. 
Qed.

(** ** Permissions on the Cminor stack block *)

(** The parts of the Cminor stack data block that are not images of
    C#minor local variable blocks remain freeable at all times. *)

Inductive is_reachable_from_env (f: meminj) (e: Csharpminor.env) (sp: block) (ofs: Z) : Prop :=
  | is_reachable_intro: forall id b sz delta,
      e!id = Some(b, sz) ->
      f b = Some(sp, delta) ->
      delta <= ofs < delta + sz ->
      is_reachable_from_env f e sp ofs.

Definition padding_freeable (f: meminj) (e: Csharpminor.env) (tm: mem) (sp: block) (sz: Z) : Prop :=
  forall ofs, 
  0 <= ofs < sz -> Mem.perm tm sp ofs Cur Freeable \/ is_reachable_from_env f e sp ofs.

Lemma padding_freeable_invariant:
  forall f1 e tm1 sp sz cenv lo hi f2 tm2,
  padding_freeable f1 e tm1 sp sz ->
  match_env f1 cenv e sp lo hi ->
  (forall ofs, Mem.perm tm1 sp ofs Cur Freeable -> Mem.perm tm2 sp ofs Cur Freeable) ->
  (forall b, b < hi -> f2 b = f1 b) ->
  padding_freeable f2 e tm2 sp sz.
Proof.
  intros; red; intros.
  exploit H; eauto. intros [A | A].
  left; auto.
  right. inv A. exploit me_bounded; eauto. intros [D E].
  econstructor; eauto. rewrite H2; auto. 
Qed.

(** Decidability of the [is_reachable_from_env] predicate. *)

Lemma is_reachable_from_env_dec:
  forall f e sp ofs, is_reachable_from_env f e sp ofs \/ ~is_reachable_from_env f e sp ofs.
Proof.
  intros. 
  set (pred := fun id_b_sz : ident * (block * Z) =>
                 match id_b_sz with
                 | (id, (b, sz)) =>
                      match f b with
                           | None => false
                           | Some(sp', delta) =>
                               if eq_block sp sp'
                               then zle delta ofs && zlt ofs (delta + sz)
                               else false
                      end
                 end).
  destruct (List.existsb pred (PTree.elements e)) eqn:?.
  (* yes *)
  rewrite List.existsb_exists in Heqb. 
  destruct Heqb as [[id [b sz]] [A B]].
  simpl in B. destruct (f b) as [[sp' delta] |] eqn:?; try discriminate.
  destruct (eq_block sp sp'); try discriminate.
  destruct (andb_prop _ _ B).
  left. apply is_reachable_intro with id b sz delta.
  apply PTree.elements_complete; auto. 
  congruence.
  split; eapply proj_sumbool_true; eauto.
  (* no *)
  right; red; intro NE; inv NE. 
  assert (existsb pred (PTree.elements e) = true).
  rewrite List.existsb_exists. exists (id, (b, sz)); split.
  apply PTree.elements_correct; auto. 
  simpl. rewrite H0. rewrite dec_eq_true.
  unfold proj_sumbool. destruct H1. rewrite zle_true; auto. rewrite zlt_true; auto. 
  congruence.
Qed.

(** * Correspondence between global environments *)

(** Global environments match if the memory injection [f] leaves unchanged
  the references to global symbols and functions. *)

Inductive match_globalenvs (f: meminj) (bound: Z): Prop :=
  | mk_match_globalenvs
      (POS: bound > 0)
      (DOMAIN: forall b, b < bound -> f b = Some(b, 0))
      (IMAGE: forall b1 b2 delta, f b1 = Some(b2, delta) -> b2 < bound -> b1 = b2)
      (SYMBOLS: forall id b, Genv.find_symbol ge id = Some b -> b < bound)
      (FUNCTIONS: forall b fd, Genv.find_funct_ptr ge b = Some fd -> b < bound)
      (VARINFOS: forall b gv, Genv.find_var_info ge b = Some gv -> b < bound).

Remark inj_preserves_globals:
  forall f hi,
  match_globalenvs f hi ->
  meminj_preserves_globals ge f.
Proof.
  intros. inv H.
  split. intros. apply DOMAIN. eapply SYMBOLS. eauto.
  split. intros. apply DOMAIN. eapply VARINFOS. eauto. 
  intros. symmetry. eapply IMAGE; eauto. 
Qed.

(** * Invariant on abstract call stacks  *)

(** Call stacks represent abstractly the execution state of the current
  C#minor and Cminor functions, as well as the states of the
  calling functions.  A call stack is a list of frames, each frame
  collecting information on the current execution state of a C#minor
  function and its Cminor translation. *)

Inductive frame : Type :=
  Frame(cenv: compilenv)
       (tf: Cminor.function)
       (e: Csharpminor.env)
       (le: Csharpminor.temp_env)
       (te: Cminor.env)
       (sp: block)
       (lo hi: Z).

Definition callstack : Type := list frame.

(** Matching of call stacks imply:
- matching of environments for each of the frames
- matching of the global environments
- separation conditions over the memory blocks allocated for C#minor local variables;
- separation conditions over the memory blocks allocated for Cminor stack data;
- freeable permissions on the parts of the Cminor stack data blocks
  that are not images of C#minor local variable blocks.
*)

Inductive match_callstack (f: meminj) (m: mem) (tm: mem):
                          callstack -> Z -> Z -> Prop :=
  | mcs_nil:
      forall hi bound tbound,
      match_globalenvs f hi ->
      hi <= bound -> hi <= tbound ->
      match_callstack f m tm nil bound tbound 
  | mcs_cons:
      forall cenv tf e le te sp lo hi cs bound tbound l'
        (BOUND: hi <= bound)
        (TBOUND: sp < tbound)
        (MTMP: match_temps f le te)
        (MENV: match_env f cenv e sp lo hi)
        (BOUND: match_bounds e m)
        (PERM: padding_freeable f e tm sp tf.(fn_stackspace))
        (VALID: valid_region_alloc tf.(fn_id) (tf.(fn_dummyblocks)))
        (SPTAG: Mem.tget tm sp = Some (Mem.Tag_stack (Mem.Stack_main tf.(fn_id) (Some tf.(fn_stackspace)))))
        (DUMMYBLOCKLIST: tf.(fn_dummyblocks) = l')
        (CENV: Cminorgen.assign_variables_aux (PTree.empty _) l' = cenv)
        (ENV: Csharpminor.assign_variables_aux (PTree.empty _) lo (List.map (fun i: (ident * (Z * Z * Z)) => match i with (id, (sz, al, of)) => (id, (sz, al, Some of))end) l') = e)
        (NOREPET: list_norepet (map fst l'))
        (*DUMMYBLOCKS:
           forall (id : positive) (ofs : Z),
             cenv ! id = Some ofs ->
             forall (b : block) (sz : Z),
               e ! id = Some (b, sz) ->
               exists al, exists n, b =  lo + Z.of_nat n /\
                         nth_error (fn_dummyblocks tf) n = Some (sz, al, ofs)*)
        (TAG: forall n id sz al ofs, List.nth_error l' n = Some (id, (sz, al, ofs)) ->
                                     Mem.tget m (lo + Z.of_nat n) = Some (Mem.Tag_stack (Mem.Stack_var (Mem.Svar tf.(fn_id) id sp sz al (Some ofs)))))
        (HI: hi = lo + Z.of_nat (length l'))
        (MCS: match_callstack f m tm cs lo sp),
      match_callstack f m tm (Frame cenv tf e le te sp lo hi :: cs) bound tbound.

(** [match_callstack] implies [match_globalenvs]. *)

Lemma match_callstack_match_globalenvs:
  forall f m tm cs bound tbound,
  match_callstack f m tm cs bound tbound ->
  exists hi, match_globalenvs f hi.
Proof.
  induction 1; eauto.
Qed.

(** Invariance properties for [match_callstack]. *)


Lemma nth_error_ex: forall {A: Type} n l, (n < length l)%nat ->
                                exists a: A, nth_error l n = Some a.
Proof with intros; eauto; try omega.
  induction n; destruct l; simpl...
  apply IHn...
Qed.  

Lemma nth_error_length: forall {A: Type} n l (a:A), nth_error l n = Some a ->
                                              (n < length l)%nat.
Proof.
  induction n; destruct l; simpl; try discriminate.
   intros; omega.
  intros. exploit IHn. eassumption. omega.
Qed.

Lemma match_callstack_invariant:
  forall f1 m1 tm1 f2 m2 tm2 cs bound tbound,
  match_callstack f1 m1 tm1 cs bound tbound ->
  inject_incr f1 f2 ->
  (forall b ofs p, b < bound -> Mem.perm m2 b ofs Max p -> Mem.perm m1 b ofs Max p) ->
  (forall sp ofs, sp < tbound -> Mem.perm tm1 sp ofs Cur Freeable -> Mem.perm tm2 sp ofs Cur Freeable) ->
  (forall b, b < bound -> f2 b = f1 b) ->
  (forall b, b < bound -> Mem.tget m2 b = Mem.tget m1 b) ->
  (forall sp, sp < tbound -> Mem.tget tm2 sp = Mem.tget tm1 sp) ->
  (forall b b' delta, f2 b = Some(b', delta) -> b' < tbound -> f1 b = Some(b', delta)) ->
  match_callstack f2 m2 tm2 cs bound tbound.
Proof.
  induction 1; intros.
  (* base case *)
  econstructor; eauto.
  inv H. constructor; intros; eauto.
  eapply IMAGE; eauto. eapply H8; eauto. omega.
  (* inductive case *)
  assert (lo <= hi) by (eapply me_low_high; eauto).
  econstructor; eauto.
  eapply match_temps_invariant; eauto.
  eapply match_env_invariant; eauto. 
    intros. apply H3. omega.
  eapply match_bounds_invariant; eauto.
    intros. eapply H1; eauto. 
    exploit me_bounded; eauto. omega. 
  eapply padding_freeable_invariant; eauto. 
    intros. apply H3. omega.
  rewrite H5. eauto. assumption.
  intros. rewrite H4. eauto. generalize (nth_error_length _ _ _ H8). intro. generalize (inj_lt _ _ H9). omega.  
  eapply IHmatch_callstack; eauto. 
    intros. eapply H1; eauto. omega. 
    intros. eapply H2; eauto. omega. 
    intros. eapply H3; eauto. omega.
    intros. eapply H4; eauto. omega.
    intros. eapply H5; eauto. omega.
    intros. eapply H6; eauto. omega.
Qed.

Lemma match_callstack_incr_bound:
  forall f m tm cs bound tbound bound' tbound',
  match_callstack f m tm cs bound tbound ->
  bound <= bound' -> tbound <= tbound' ->
  match_callstack f m tm cs bound' tbound'.
Proof.
  intros. inv H.
  econstructor; eauto. omega. omega. 
  econstructor; eauto. omega. omega.
Qed.

(** Assigning a temporary variable. *)

Lemma match_callstack_set_temp:
  forall f cenv e le te sp lo hi cs bound tbound m tm tf id v tv,
  val_inject f v tv ->
  match_callstack f m tm (Frame cenv tf e le te sp lo hi :: cs) bound tbound ->
  match_callstack f m tm (Frame cenv tf e (PTree.set id v le) (PTree.set id tv te) sp lo hi :: cs) bound tbound.
Proof.
  intros. inv H0. econstructor; eauto.
  eapply match_temps_assign; eauto. 
Qed.

(** Preservation of [match_callstack] by freeing all blocks allocated
  for local variables at function entry (on the C#minor side)
  and simultaneously freeing the Cminor stack data block. *)

Lemma map_snd_commut: forall x,
      map snd
          (map
             (fun i : ident * (Z * Z * Z) =>
                let (id, y) := i in
                let (y0, al) := y in
                let (sz, of) := y0 in (id, (sz, of, Some al))) x) = 
      map (fun i => match i with (sz, of, al) => (sz, of, Some al) end) (map snd x).
Proof.
 induction x; simpl.
    reflexivity.
   destruct a; simpl.
   destruct p; simpl.
   destruct p; simpl.
   congruence.
Qed.

Lemma list_norepet_nth_error:
  forall {A: Type} l,
    list_norepet l ->
    forall n1 (a: A), List.nth_error l n1 = Some a ->
      forall n2, List.nth_error l n2 = Some a ->
                 n1 = n2.
Proof.
  induction 1; simpl.
   destruct n1; destruct n2; discriminate.
  destruct n1; simpl.
   injection 1; intros; subst.
   destruct n2; simpl in *.
    reflexivity.
   destruct H. eauto using nth_error_in.
  intros. destruct n2; simpl in *.
  inv H2. destruct H. eauto using nth_error_in.
  f_equal. eauto.
Qed.

Lemma nth_error_map:
  forall {A B: Type} (f: A -> B) n l a,
    nth_error (map f l) n = Some a ->
    exists b, nth_error l n = Some b /\ f b = a.
Proof.
  intros until a. rewrite list_map_nth. destruct (nth_error l n); try discriminate. simpl. injection 1; intros; subst; eauto.
Qed.

Lemma csharpminor_assign_variables_aux_prop: forall l e b id b' sz,
(Csharpminor.assign_variables_aux e b l) ! id = Some (b', sz) ->
 e ! id = Some (b', sz) \/
 exists n, exists al, exists of, List.nth_error l n = Some (id, (sz, al, of)) /\ b' = b + Z.of_nat n.
Proof.
  induction l; simpl.
   tauto.
  destruct a. destruct p. destruct p. intros.
  destruct (IHl _ _ _ _ _ H).
  destruct (peq i id).
   subst. rewrite PTree.gss in H0. inv H0.
   right. exists O. esplit. esplit. split. reflexivity. Transparent Z.of_nat. simpl; unfold block; omega.
  rewrite PTree.gso in H0. tauto. auto.
 destruct H0. destruct H0. destruct H0. destruct H0. subst.
 right. exists (S x). esplit. esplit. split. eassumption. rewrite Nat2Z.inj_succ. unfold Z.succ. unfold block; omega.
Qed.

Lemma cminorgen_assign_variables_aux_prop:
forall l e id ofs,
  (Cminorgen.assign_variables_aux e l) ! id = Some ofs ->
  e ! id = Some ofs \/
  exists n, exists sz, exists al,
                         List.nth_error l n = Some (id, (sz, al, ofs)).
Proof.
  intros until e.
  functional induction (Cminorgen.assign_variables_aux e l).
   tauto.
  intros.
  exploit IHc; eauto. clear IHc.
  destruct (peq id0 id).
   subst. rewrite PTree.gss. destruct 1.
    inv H0. right. exists O. esplit. esplit. reflexivity.
   destruct H0. destruct H0. destruct H0. right. eexists (S _). esplit. esplit. eassumption.
  rewrite PTree.gso. destruct 1.
   tauto.
  destruct H0. destruct H0. destruct H0. right. eexists (S _). esplit. esplit. eassumption.
  assumption.
Qed.  

Lemma Z_of_nat_le_inv:
 forall n x, 0 <= x < Z.of_nat n -> exists p, x = Z.of_nat p /\ (p < n)%nat.
Proof.
  induction n.
   simpl. intros; exfalso; omega.
  intro. rewrite Nat2Z.inj_succ.
  intro.
  destruct (zeq x (Z.of_nat n)).
   exists n. eauto.
  exploit (IHn x).
   omega.
  destruct 1.
  destruct H0.
  eauto.
Qed.

Lemma csharpminor_assign_variables_aux_none:
  forall l e b id,
    (Csharpminor.assign_variables_aux e b l) ! id = None ->
    e ! id = None /\
    (forall sao, In (id, sao) l -> False).
Proof.
  intros until b.
  functional induction (Csharpminor.assign_variables_aux e b l).
   simpl; tauto.
  intros. exploit IHe0. eassumption. destruct 1.
  destruct (peq id0 id).
   subst. rewrite PTree.gss in H0. discriminate.
  rewrite PTree.gso in H0.
  split; simpl; auto.
  destruct 1; eauto.
  congruence.
  assumption.
Qed.

Lemma cminorgen_assign_variables_aux_none:
  forall l e id,
    (Cminorgen.assign_variables_aux e l) ! id = None ->
    e ! id = None /\
    (forall sao, In (id, sao) l -> False).
Proof.
  intros until e. functional induction (Cminorgen.assign_variables_aux e l).
   simpl; tauto.
  intros. exploit IHc. eassumption. destruct 1.
  destruct (peq id0 id).
   subst. rewrite PTree.gss in H0. discriminate.
  rewrite PTree.gso in H0.
  split; simpl; auto.
  destruct 1; eauto.
  congruence.
  assumption.
Qed.

Lemma region_free_perm_recip:
  forall l lo m m',
    Mem.region_free lo m l = Some m' ->
    forall b ofs k p,
      Mem.perm m' b ofs k p ->
      Mem.perm m b ofs k p.
Proof.
  intros until m.
  functional induction (Mem.region_free lo m l); try discriminate.
   congruence.
  intros. eapply Mem.perm_free_3. eassumption. eauto.
Qed.

Lemma region_free_tget:
  forall l lo m m',
    Mem.region_free lo m l = Some m' ->
    forall b, Mem.tget m' b = Mem.tget m b.
Proof.
  intros until m.
  functional induction (Mem.region_free lo m l); try discriminate.
   congruence.
  intros. erewrite IHo. 2: eassumption. eapply Mem.tget_free. eassumption.
Qed.

Lemma region_free_nextblock:
  forall l lo m m',
    Mem.region_free lo m l = Some m' ->
    Mem.nextblock m' = Mem.nextblock m.
Proof.
  intros until m.
  functional induction (Mem.region_free lo m l); try discriminate.
   congruence.
  intros. rewrite IHo. eapply Mem.nextblock_free. eassumption. assumption.
Qed.

Lemma match_callstack_region_free:
  forall f cenv tf e le te sp lo hi cs m m' tm,
  Mem.inject f m tm ->
  Mem.region_free lo m (map snd tf.(fn_dummyblocks)) = Some m' ->
  match_callstack f m tm (Frame cenv tf e le te sp lo hi :: cs) (Mem.nextblock m) (Mem.nextblock tm) ->
  exists tm',
  Mem.free tm sp 0 tf.(fn_stackspace) = Some tm'
  /\ Mem.inject f m' tm'
  /\ match_callstack f m' tm' cs (Mem.nextblock m') (Mem.nextblock tm').
Proof.
  inversion 3; subst.
  inversion MENV; subst.
  exploit Mem.free_region_inject.
  eassumption.
  eassumption.
  erewrite Mem.tget_inject. eassumption.
  eassumption.
  replace (lo - 1 + 1) with lo by (unfold block; omega).
  intros.  eauto.
  replace (lo - 1 + 1) with lo by (unfold block; omega).
  eassumption.
 destruct 1.
 destruct H2.
 esplit.
 split.
 eassumption.
 split.
 assumption.
 eapply match_callstack_incr_bound.
 eapply match_callstack_invariant.
 eassumption.
 apply inject_incr_refl.
 intros. 
 eauto using region_free_perm_recip.
 intros. eapply Mem.perm_free_1. eassumption. left. unfold block; omega. assumption.
 trivial.
 intros. eapply region_free_tget. eassumption.
 intros. eapply Mem.tget_free. eassumption.
 trivial.
 erewrite region_free_nextblock. 2: eassumption. omega.
 erewrite Mem.nextblock_free. 2: eassumption. omega.
Qed.

(** Preservation of [match_callstack] by external calls. *)

Lemma match_callstack_external_call:
  forall f1 f2 m1 m2 m1' m2',
  mem_unchanged_on (loc_unmapped f1) m1 m2 ->
  mem_unchanged_on (loc_out_of_reach f1 m1) m1' m2' ->
  inject_incr f1 f2 ->
  inject_separated f1 f2 m1 m1' ->
  (forall b ofs p, Mem.valid_block m1 b -> Mem.perm m2 b ofs Max p -> Mem.perm m1 b ofs Max p) ->
  forall TAG: (forall b, Mem.valid_block m1 b -> Mem.tget m2 b = Mem.tget m1 b),
  forall TAG': (forall sp, Mem.valid_block m1' sp -> Mem.tget m2' sp = Mem.tget m1' sp),
  forall cs bound tbound,
  match_callstack f1 m1 m1' cs bound tbound ->
  bound <= Mem.nextblock m1 -> tbound <= Mem.nextblock m1' ->
  match_callstack f2 m2 m2' cs bound tbound.
Proof.
  intros until m2'. 
  intros UNMAPPED OUTOFREACH INCR SEPARATED MAXPERMS TAG TAG'.
  destruct OUTOFREACH as [OUTOFREACH1 OUTOFREACH2].
  induction 1; intros.
(* base case *)
  apply mcs_nil with hi; auto.
  inv H. constructor; auto.
  intros. case_eq (f1 b1). 
  intros [b2' delta'] EQ. rewrite (INCR _ _ _ EQ) in H. inv H. eauto. 
  intro EQ. exploit SEPARATED; eauto. intros [A B]. elim B. red. omega. 
(* inductive case *)
  econstructor. auto. auto. 
  eapply match_temps_invariant; eauto.
  eapply match_env_invariant; eauto. 
  red in SEPARATED. intros. destruct (f1 b) as [[b' delta']|] eqn:?. 
  exploit INCR; eauto. congruence.
  exploit SEPARATED; eauto. intros [A B]. elim B. red. omega. 
  intros. assert (lo <= hi) by (eapply me_low_high; eauto).
  destruct (f1 b) as [[b' delta']|] eqn:?. 
  apply INCR; auto. 
  destruct (f2 b) as [[b' delta']|] eqn:?; auto.
  exploit SEPARATED; eauto. intros [A B]. elim A. red. omega.
  eapply match_bounds_invariant; eauto. 
  intros. eapply MAXPERMS; eauto. red. exploit me_bounded; eauto. omega. 
  (* padding-freeable *)
  red; intros.
  destruct (is_reachable_from_env_dec f1 e sp ofs).
  inv H3. right. apply is_reachable_intro with id b sz delta; auto. 
  exploit PERM; eauto. intros [A|A]; try contradiction.
  left. apply OUTOFREACH1; auto. red; intros.
  red; intros; elim H3.
  exploit me_inv; eauto. intros [id [lv B]].
  exploit BOUND0; eauto. intros C.
  apply is_reachable_intro with id b0 lv delta; auto; omega.
  (* valid *)
  assumption.
  (* tag main *)
  rewrite TAG'. auto. unfold Mem.valid_block. omega.
  (* dummyblocks *)
  eassumption.
  (* cenv *)
  assumption.  
  (* env *)
  assumption.
  (* norepet *)
  assumption.
  (* tag var *)
  intros. rewrite TAG. eauto. exploit TAG0; eauto. intro. eapply Mem.tget_valid. congruence.
  (* hi *)
  assumption.
  (* induction *)
  eapply IHmatch_callstack; eauto. inv MENV; omega. omega.
Qed.

(** [match_callstack] and allocations *)

(** Initialization of C#minor temporaries and Cminor local variables. *)

Lemma create_undef_temps_val:
  forall id v temps, (create_undef_temps temps)!id = Some v -> In id temps /\ v = Vundef.
Proof.
  induction temps; simpl; intros.
  rewrite PTree.gempty in H. congruence.
  rewrite PTree.gsspec in H. destruct (peq id a).
  split. auto. congruence.
  exploit IHtemps; eauto. tauto. 
Qed.

Fixpoint set_params' (vl: list val) (il: list ident) (te: Cminor.env) : Cminor.env :=
  match il, vl with
  | i1 :: is, v1 :: vs => set_params' vs is (PTree.set i1 v1 te)
  | i1 :: is, nil => set_params' nil is (PTree.set i1 Vundef te)
  | _, _ => te
  end.


Lemma bind_parameters_agree_rec:
  forall f vars vals tvals le1 le2 te,
  bind_parameters vars vals le1 = Some le2 ->
  val_list_inject f vals tvals ->
  match_temps f le1 te ->
  match_temps f le2 (set_params' tvals vars te).
Proof.
Opaque PTree.set.
  induction vars; simpl; intros.
  destruct vals; try discriminate. inv H. auto. 
  destruct vals; try discriminate. inv H0.
  simpl. eapply IHvars; eauto.
  red; intros. rewrite PTree.gsspec in *. destruct (peq id a). 
  inv H0. exists v'; auto.
  apply H1; auto.
Qed.

Lemma set_params'_outside:
  forall id il vl te, ~In id il -> (set_params' vl il te)!id = te!id.
Proof.
  induction il; simpl; intros. auto. 
  destruct vl; rewrite IHil.
  apply PTree.gso. intuition. intuition.
  apply PTree.gso. intuition. intuition.
Qed.

Lemma set_params'_inside:
  forall id il vl te1 te2,
  In id il ->
  (set_params' vl il te1)!id = (set_params' vl il te2)!id.
Proof.
  induction il; simpl; intros.
  contradiction.
  destruct vl; destruct (List.in_dec peq id il); auto;
  repeat rewrite set_params'_outside; auto;
  assert (a = id) by intuition; subst a; repeat rewrite PTree.gss; auto.
Qed.

Lemma set_params_set_params':
  forall il vl id,
  list_norepet il ->
  (set_params vl il)!id = (set_params' vl il (PTree.empty val))!id.
Proof.
  induction il; simpl; intros.
  auto.
  inv H. destruct vl. 
  rewrite PTree.gsspec. destruct (peq id a). 
  subst a. rewrite set_params'_outside; auto. rewrite PTree.gss; auto.
  rewrite IHil; auto.
  destruct (List.in_dec peq id il). apply set_params'_inside; auto.
  repeat rewrite set_params'_outside; auto. rewrite PTree.gso; auto. 
  rewrite PTree.gsspec. destruct (peq id a). 
  subst a. rewrite set_params'_outside; auto. rewrite PTree.gss; auto.
  rewrite IHil; auto.
  destruct (List.in_dec peq id il). apply set_params'_inside; auto.
  repeat rewrite set_params'_outside; auto. rewrite PTree.gso; auto. 
Qed.

Lemma set_locals_outside:
  forall e id il,
  ~In id il -> (set_locals il e)!id = e!id.
Proof.
  induction il; simpl; intros.
  auto.
  rewrite PTree.gso. apply IHil. tauto. intuition. 
Qed.

Lemma set_locals_inside:
  forall e id il,
  In id il -> (set_locals il e)!id = Some Vundef.
Proof.
  induction il; simpl; intros.
  contradiction.
  destruct H. subst a. apply PTree.gss. 
  rewrite PTree.gsspec. destruct (peq id a). auto. auto. 
Qed.

Lemma set_locals_set_params':
  forall vars vals params id,
  list_norepet params ->
  list_disjoint params vars ->
  (set_locals vars (set_params vals params)) ! id =
  (set_params' vals params (set_locals vars (PTree.empty val))) ! id.
Proof.
  intros. destruct (in_dec peq id vars).
  assert (~In id params). apply list_disjoint_notin with vars; auto. apply list_disjoint_sym; auto.
  rewrite set_locals_inside; auto. rewrite set_params'_outside; auto. rewrite set_locals_inside; auto.
  rewrite set_locals_outside; auto. rewrite set_params_set_params'; auto. 
  destruct (in_dec peq id params). 
  apply set_params'_inside; auto.
  repeat rewrite set_params'_outside; auto. 
  rewrite set_locals_outside; auto. 
Qed.

Lemma bind_parameters_agree:
  forall f params temps vals tvals le,
  bind_parameters params vals (create_undef_temps temps) = Some le ->
  val_list_inject f vals tvals ->
  list_norepet params ->
  list_disjoint params temps ->
  match_temps f le (set_locals temps (set_params tvals params)).
Proof.
  intros; red; intros.
  exploit bind_parameters_agree_rec; eauto.
  instantiate (1 := set_locals temps (PTree.empty val)).
  red; intros. exploit create_undef_temps_val; eauto. intros [A B]. subst v0.
  exists Vundef; split. apply set_locals_inside; auto. auto.
  intros [v' [A B]]. exists v'; split; auto.
  rewrite <- A. apply set_locals_set_params'; auto.
Qed.

Lemma region_alloc_perm_old:
  forall m b ofs k p,
    Mem.perm m b ofs k p ->
    forall fi l b0,
      Mem.perm (Mem.region_alloc b0 fi m l) b ofs k p.
Proof.
  intros until l. revert m b ofs k p H. induction l; simpl.
   tauto.
  destruct a.
  destruct p.
  destruct p.
  intros.
  case_eq (Mem.alloc m 0 z0 (Mem.Tag_stack (Mem.Stack_var (Mem.Svar fi i b0 z0 z1 (Some z))))).
  intros.
  eapply IHl.
  eapply Mem.perm_alloc_1.
  eassumption.
  assumption.
Qed.

Lemma region_alloc_perm_old_recip:
  forall l fi b0 m b ofs k p,
    Mem.perm (Mem.region_alloc b0 fi m l) b ofs k p ->
    Mem.valid_block m b ->
    Mem.perm m b ofs k p.
Proof.
  induction l; simpl.
   tauto.
  intros until p.
  destruct a. destruct p0. destruct p0. case_eq (Mem.alloc m 0 z0 (Mem.Tag_stack (Mem.Stack_var (Mem.Svar fi i b0 z0 z1 (Some z))))). intros. eapply Mem.perm_alloc_4. eassumption. eapply IHl. eassumption. eapply Mem.valid_block_alloc. eassumption. assumption.
  unfold Mem.valid_block in H1. generalize (Mem.alloc_result _ _ _ _ _ _ H). unfold block in *. omega.
Qed.

Lemma region_alloc_perm_new:
  forall n l i sz al of,
    List.nth_error l n = Some (i, (sz, al, of)) ->
    forall fi b0 m ofs k p,
      Mem.perm (Mem.region_alloc b0 fi m l) (Mem.nextblock m + Z.of_nat n) ofs k p ->
      0 <= ofs < sz.
Proof.
  Opaque Z.of_nat.
  induction n; simpl.
   destruct l; try discriminate.
   injection 1; intros until 1; subst. clear H.
   simpl. intros until p. case_eq (Mem.alloc m 0 sz (Mem.Tag_stack (Mem.Stack_var (Mem.Svar fi i b0 sz al (Some of))))). intros. exploit region_alloc_perm_old_recip. eassumption. generalize (Mem.alloc_result _ _ _ _ _ _ H). unfold Mem.valid_block. generalize (Mem.nextblock_alloc _ _ _ _ _ _ H). unfold Z.succ, block in *. Transparent Z.of_nat. simpl; omega.
   generalize (Mem.alloc_result _ _ _ _ _ _ H). intro; subst. replace (Mem.nextblock m + 0) with (Mem.nextblock m) by (unfold block in *; omega). intro. eapply Mem.perm_alloc_3. eassumption. eassumption.
  destruct l; try discriminate.
  rewrite Nat2Z.inj_succ.
  Opaque Z.of_nat.
  simpl.
  intros until p0.  
  destruct p. destruct p. destruct p. case_eq (Mem.alloc m 0 z0 (Mem.Tag_stack (Mem.Stack_var (Mem.Svar fi i0 b0 z0 z1 (Some z))))). replace (Mem.nextblock m + Z.succ (Z.of_nat n)) with (Z.succ (Mem.nextblock m) + Z.of_nat n).
  intros until 1. rewrite <- (Mem.nextblock_alloc _ _ _ _ _ _ H0).
 intros. eapply IHn. eassumption. eassumption.
 unfold Z.succ. omega.
Qed.

Lemma alloc_dummyblocks_perm_recip:
  forall l b fi m b' ofs k p,
    Mem.perm (Mem.alloc_dummyblocks b fi m l) b' ofs k p ->
    Mem.perm m b' ofs k p.
Proof.
  intros until m. functional induction (Mem.alloc_dummyblocks b fi m l).
   tauto.
  intros. eapply Mem.perm_alloc_4. eassumption. eauto. intro. subst. exploit Mem.perm_alloc_3. eassumption. eapply IHt. eassumption. omega.
Qed.

Lemma alloc_dummyblocks_perm:
  forall m b' fi ofs k p,
    Mem.perm m b' ofs k p ->
    forall l b,
      Mem.perm (Mem.alloc_dummyblocks b fi m l) b' ofs k p.
Proof.
  intros. revert b' ofs k p H. functional induction (Mem.alloc_dummyblocks b fi m l).
   tauto.
  intros. eapply IHt. eapply Mem.perm_alloc_1. eassumption. assumption.
Qed.

Lemma alloc_dummyblocks_tget: forall b m,
Mem.valid_block m b ->
forall fi b' l,
Mem.tget (Mem.alloc_dummyblocks b' fi m l) b = Mem.tget m b.
Proof.
  intros. revert b H. functional induction (Mem.alloc_dummyblocks b' fi m l).
   reflexivity.
  intros. rewrite IHt. eapply Mem.tget_alloc_other. eassumption. intro; subst. eapply Mem.fresh_block_alloc. eassumption. assumption.
  eapply Mem.valid_block_alloc. eassumption. assumption.
Qed.

Lemma nextblock_alloc_dummyblocks:
  forall b' fi m l,
    Mem.nextblock (Mem.alloc_dummyblocks b' fi m l) = Mem.nextblock m + Z.of_nat (length l).
Proof.
  Opaque Z.of_nat.
  intros. functional induction (Mem.alloc_dummyblocks b' fi m l); simpl.
   Transparent Z.of_nat. simpl; unfold block; omega.
  rewrite Nat2Z.inj_succ. Opaque Z.of_nat. rewrite IHt.
  erewrite Mem.nextblock_alloc. 2: eassumption.
  unfold block; omega.
Qed.

Corollary nextblock_alloc_dummyblocks_le:
  forall b' fi m l,
    Mem.nextblock m <= Mem.nextblock (Mem.alloc_dummyblocks b' fi m l).
Proof.
  intros. rewrite nextblock_alloc_dummyblocks. 
  generalize (Zle_0_nat (length l)). omega.
Qed.


Lemma nextblock_region_alloc:
  forall b' fi m l,
    Mem.nextblock (Mem.region_alloc b' fi m l) = Mem.nextblock m + Z.of_nat (length l).
Proof.
  Opaque Z.of_nat.
  intros. functional induction (Mem.region_alloc b' fi m l); simpl.
   Transparent Z.of_nat. simpl; unfold block; omega.
  rewrite Nat2Z.inj_succ. Opaque Z.of_nat. rewrite IHt.
  erewrite Mem.nextblock_alloc. 2: eassumption.
  unfold block; omega.
Qed.

Corollary nextblock_region_alloc_le:
  forall b' fi m l,
    Mem.nextblock m <= Mem.nextblock (Mem.region_alloc b' fi m l).
Proof.
  intros. rewrite nextblock_region_alloc.
  generalize (Zle_0_nat (length l)). omega.
Qed.

Lemma tget_region_alloc_other: forall b fi m, Mem.valid_block m b -> forall b' l,
  Mem.tget (Mem.region_alloc b' fi m l) b = Mem.tget m b.
Proof.
  intros. revert b H. functional induction (Mem.region_alloc b' fi m l).
   reflexivity.
  intros. rewrite IHt. eapply Mem.tget_alloc_other. eassumption. intro. subst. eapply Mem.fresh_block_alloc. eassumption. assumption.
  eapply Mem.valid_block_alloc. eassumption. assumption.
Qed.

Lemma tget_region_alloc_same: forall n l i sz al of,
  nth_error l n = Some (i, (sz, al, of)) ->
  forall b' fi m,
  Mem.tget (Mem.region_alloc b' fi m l) (Mem.nextblock m + Z.of_nat n) = Some (Mem.Tag_stack (Mem.Stack_var (Mem.Svar fi i b' sz al (Some of)))).
Proof.
  Opaque Z.of_nat.
  induction n; destruct l; try discriminate; simpl.
   injection 1; intros; subst.
   case_eq (Mem.alloc m 0 sz (Mem.Tag_stack (Mem.Stack_var (Mem.Svar fi i b' sz al (Some of))))).
   intros.
   replace (Mem.nextblock m + Z.of_nat 0) with (Mem.nextblock m).
   generalize (Mem.alloc_result _ _ _ _ _ _ H0). intro; subst.
   rewrite tget_region_alloc_other.
   eapply Mem.tget_alloc_same. eassumption.
   eapply Mem.valid_new_block. eassumption.
   Transparent Z.of_nat. simpl; unfold block; omega.
  destruct p. destruct p. destruct p.
  intros.
  case_eq (Mem.alloc m 0 z0 (Mem.Tag_stack (Mem.Stack_var (Mem.Svar fi i b' z0 z1 (Some z))))).
  intros.
  generalize (Mem.alloc_result _ _ _ _ _ _ H0). intro; subst.
  generalize (Mem.nextblock_alloc _ _ _ _ _ _ H0). intro.
  rewrite Nat2Z.inj_succ. replace (Mem.nextblock m + Z.succ (Z.of_nat n)) with (Mem.nextblock m0 + Z.of_nat n). erewrite IHn. reflexivity. assumption.
  unfold block in *; omega.
Qed.

Theorem map_fst_commut: forall l',
                          map fst (
                                map
                                  (fun i : ident * (Z * Z * Z) =>
                                     let (id, y) := i in
                                     let (y0, al) := y in
                                     let (sz, of) := y0 in (id, (sz, of, Some al))) l') =
                          map fst l'.
Proof.
  induction l'; simpl.
   reflexivity.
  destruct a. destruct p. destruct p. simpl. congruence.
Qed.
                                  
Theorem match_callstack_function_entry:
  forall fb fn tf cenv l' m m0 e m' tm b f cs args targs le osz
         (HSYMB: Genv.find_funct_ptr ge fb = Some (Internal fn))
         (HTRANS: transl_function_aux fn = OK ((cenv, l'), tf))
         (HSIZE: stack_size (Csharpminor.fn_vars fn) <= Int.max_unsigned)
         (HVARS_NOREPET: list_norepet (map fst (Csharpminor.fn_vars fn)))
         (HPARAMS_NOREPET: list_norepet (Csharpminor.fn_params fn))
         (HPARAMS_TEMPS_DISJOINT: list_disjoint (Csharpminor.fn_params fn) (Csharpminor.fn_temps fn))
         (HCALL_COND: call_condition true (Csharpminor.fn_vars fn))
         (HALIGN: forall (id : ident) (sz al : Z) (of : option Z),
          In (id, (sz, al, of)) (Csharpminor.fn_vars fn) ->
          (0 <= sz /\ 0 < al /\ (al | sz) /\ (al | 8)))
         (HALLOC_DUMMY: Mem.alloc m 0 0
                                  (Mem.Tag_stack (Mem.Stack_main (Csharpminor.fn_id fn) osz)) =
                        (m0, b))
         (HALLOC_VARS: alloc_variables true m0 b fn.(Csharpminor.fn_id) (Csharpminor.fn_vars fn) e m' osz)
         (HBIND_PARAMS: bind_parameters (Csharpminor.fn_params fn) args (create_undef_temps fn.(fn_temps)) = Some le)
         (HARGS: val_list_inject f args targs)
         (HMATCH: match_callstack f m tm cs (Mem.nextblock m) (Mem.nextblock tm))
         (HINJ: Mem.inject f m tm),
    let te := set_locals (Csharpminor.fn_temps fn) (set_params targs (Csharpminor.fn_params fn)) in
    exists (f' : meminj) tm2 sp,
      Mem.alloc tm 0 (stack_size (Csharpminor.fn_vars fn))
                (Mem.Tag_stack
                   (Mem.Stack_main (Csharpminor.fn_id fn)
                                   (Some (stack_size (Csharpminor.fn_vars fn))))) = 
      (tm2, sp) /\
      let tm' := Mem.alloc_dummyblocks sp (fn.(Csharpminor.fn_id)) tm2 ( l') in
      Mem.inject f' m' tm' /\
      inject_incr f f' /\      
      match_callstack f' m' tm'
                      (Frame cenv tf e le te sp (Mem.nextblock m0) (Mem.nextblock m') :: cs)
                      (Mem.nextblock m') (Mem.nextblock tm')
.
Proof.
  intros.
  inv HALLOC_VARS; try discriminate.
  destruct HCALL_COND.
  destruct H0.
  generalize (check_offsets_correct _ _ Hl').
  intro VARS.  
  generalize HTRANS. intro HTRANS'.
  unfold transl_function_aux in HTRANS'. destruct (zle (stack_size (Csharpminor.fn_vars fn)) Int.max_unsigned); try discriminate. monadInv HTRANS'.
  simpl in *.
  assert (l'0 = l') by congruence.
  subst.
  exploit Mem.alloc_region_inject.
  eassumption.
  eassumption.
  eleft. eassumption. reflexivity. eassumption.
  intros.
  apply (in_map (fun i: ident * (Z * Z * Z) => let (id, y) := i in let (y0, al) := y in let (sz, of) := y0 in (id, (sz, of, Some al)))) in H2. rewrite <- VARS in H2. exploit H1; eauto. exploit HALIGN; eauto. tauto.
  intros. generalize (map_nth_error (fun i: ident * (Z * Z * Z) => let (id, y) := i in let (y0, al) := y in let (sz, of) := y0 in (id, (sz, of, Some al))) _ _ H2).   generalize (map_nth_error (fun i: ident * (Z * Z * Z) => let (id, y) := i in let (y0, al) := y in let (sz, of) := y0 in (id, (sz, of, Some al))) _ _ H3). rewrite <- VARS. intros.
  destruct (peq id1 id2).
   subst.
   destruct H4.
   generalize (map_nth_error fst _ _ H5). generalize (map_nth_error fst _ _ H6). simpl. intros.
   eauto using list_norepet_nth_error.
  generalize (nth_error_in _ _ H5). generalize (nth_error_in _ _ H6).  intros. eauto.
  rewrite <- stack_size_region_size. congruence.
 destruct 1. destruct H2.  destruct H2. destruct H2. destruct H3. destruct H4. destruct H5. destruct H6. simpl.
 esplit. esplit. esplit. split.
 eassumption.
 split. eassumption.
 split. assumption.
 econstructor.
 omega.
 (* sp valid *)
 generalize (nextblock_alloc_dummyblocks_le x2 (Csharpminor.fn_id fn) x1 (l')).
 exploit Mem.valid_new_block. eexact H2. unfold Mem.valid_block. omega.
 (* match_temps *)
 eapply bind_parameters_agree.
 eassumption.
 (* arguments *)
 eapply val_list_inject_incr. eassumption. assumption.
 (* norepet *)
 assumption. assumption.
 (* match env *)
 constructor.
 (* match var *)
 intro.
  case_eq ((Csharpminor.assign_variables_aux empty_env (Z.succ (Mem.nextblock m0 - 1)) (Csharpminor.fn_vars fn)) ! id).
   intros.
   destruct p.
   destruct (csharpminor_assign_variables_aux_prop _ _ _ _ _ _ H8).
    rewrite PTree.gempty in H9. discriminate.
   destruct H9. destruct H9. destruct H9. destruct H9. subst.
   case_eq ((assign_variables_aux (PTree.empty Z) l') ! id).
    intros.
    destruct (cminorgen_assign_variables_aux_prop _ _ _ _ H10).
     rewrite PTree.gempty in H11.
      discriminate.
     destruct H11.
     destruct H11.
     destruct H11.
     generalize (map_nth_error 
                   (fun i : ident * (Z * Z * Z) =>
                      let (id, y) := i in
                      let (y0, al) := y in
                      let (sz, of) := y0 in (id, (sz, of, Some al))) _ _ H11).
     rewrite <- VARS. intro.
     generalize (map_nth_error fst _ _ H12). generalize (map_nth_error fst _ _ H9). simpl. 
     intros. generalize (list_norepet_nth_error _ HVARS_NOREPET _ _ H14 _ H13).
     intro; subst.
     rewrite H9 in H12. inv H12.
     econstructor. econstructor. eapply H6. eassumption. rewrite Int.add_commut. rewrite Int.add_zero. reflexivity.
    intro. exfalso. destruct (cminorgen_assign_variables_aux_none _ _ _ H10).
    rewrite VARS in H9. destruct (nth_error_map _ _ _ _ H9). destruct H13. destruct x6. destruct p. destruct p. inv H14.
    eapply H12. eapply nth_error_in. eassumption.
   intros.
   case_eq ((assign_variables_aux (PTree.empty Z) l') ! id).
    intros. exfalso. destruct (cminorgen_assign_variables_aux_prop _ _ _ _ H9).
     rewrite PTree.gempty in H10. discriminate.
    destruct H10. destruct H10. destruct H10.
    generalize (map_nth_error  (fun i : ident * (Z * Z * Z) =>
            let (id, y) := i in
            let (y0, al) := y in
            let (sz, of) := y0 in (id, (sz, of, Some al))) _ _ H10). rewrite <- VARS. intro.
    destruct (csharpminor_assign_variables_aux_none _ _ _ _ H8).
    eapply H13. eapply nth_error_in. eassumption.
    intros. constructor.
  (* nextblock bound *)
  apply nextblock_region_alloc_le.
  (* region block in bounds *)
  intros. destruct (csharpminor_assign_variables_aux_prop _ _ _ _ _ _ H8).
   rewrite PTree.gempty in H9. discriminate.
  destruct H9. destruct H9. destruct H9. destruct H9.
  subst.
  generalize (nth_error_length _ _ _ H9).
  rewrite VARS. rewrite map_length. intro.
  generalize (inj_lt _ _ H10).
  intro.
  rewrite nextblock_region_alloc. omega.
  (* map blocks into stackframe *)
  intros.
  case_eq (zle (Mem.nextblock m0 - 1 + 1) b && zlt b (Mem.nextblock m0 - 1 + 1 + Z.of_nat (length ( l')))).
   intros. destruct (zle (Mem.nextblock m0 - 1 + 1) b); try discriminate. destruct (zlt b (Mem.nextblock m0 - 1 + 1 + Z.of_nat (length  l'))); try discriminate. clear H9.
   assert (0 <= b - (Mem.nextblock m0 - 1 + 1) < Z.of_nat (length l')) by omega.
   destruct (Z_of_nat_le_inv _ _ H9).
   destruct H10.
   assert (b = Mem.nextblock m0 - 1 + 1 + Z.of_nat x3) by (unfold block; omega).
   subst.
   destruct (nth_error_ex _ _ H11).
   destruct x4. destruct p. destruct p.
   generalize (H6 _ _ _ _ _ H12).
   intro. rewrite H8 in H13. inv H13.
   exists i.   
   generalize (map_nth_error  (fun i : ident * (Z * Z * Z) =>
            let (id, y) := i in
            let (y0, al) := y in
            let (sz, of) := y0 in (id, (sz, of, Some al))) _ _ H12).
    rewrite <- VARS.
   intro.
   case_eq
     ((Csharpminor.assign_variables_aux empty_env
        (Z.succ (Mem.nextblock m0 - 1)) (Csharpminor.fn_vars fn)) ! i
     ).
    intros.
    destruct p.
    destruct (csharpminor_assign_variables_aux_prop _ _ _ _ _ _ H14).
     rewrite PTree.gempty in H15. discriminate.
    destruct H15. destruct H15. destruct H15. destruct H15. subst.
    generalize (map_nth_error fst _ _ H13). generalize (map_nth_error fst _ _ H15). simpl. intros.
    generalize (list_norepet_nth_error _ HVARS_NOREPET _ _ H16 _ H17).
    intro; subst.
    esplit.
    reflexivity.
   intros.
   edestruct csharpminor_assign_variables_aux_none.
   eassumption.
   edestruct H16.
   eapply nth_error_in.
   eassumption.
  intros.
  rewrite H7 in H8.
  exploit Mem.valid_block_inject_2.
  eassumption.
  eassumption.
  intro.
  edestruct Mem.fresh_block_alloc.
  eassumption.
  assumption.
  intro. 
  destruct (zle (Mem.nextblock m0 - 1 + 1) b); try omega.
   destruct (zlt b (Mem.nextblock m0 - 1 + 1 + Z.of_nat (length ( l')))); try omega.
   discriminate.
 (* injections of blocks before *)
 intros. rewrite H7 in H8. generalize (Mem.alloc_result _ _ _ _ _ _ H2). intro; subst. eapply Mem.valid_block_inject_2. eassumption. eassumption.
 omega.
 generalize (Mem.alloc_result _ _ _ _ _ _ H2). generalize (Mem.nextblock_alloc _ _ _ _ _ _ HALLOC_DUMMY). generalize (Mem.inject_nextblock _ _ _ HINJ). unfold block; omega.
 (* match bounds *)
 unfold match_bounds. intros.
 destruct (csharpminor_assign_variables_aux_prop _ _ _ _ _ _ H8).
  rewrite PTree.gempty in H10. discriminate.
 destruct H10. destruct H10. destruct H10. destruct H10. subst.
 replace (Z.succ (Mem.nextblock m0 - 1)) with (Mem.nextblock m0) in H9.
 rewrite VARS in H10. 
 destruct (nth_error_map _ _ _ _ H10).
 destruct H11.
 destruct x6. destruct p0. destruct p0. inv H12.
 eapply region_alloc_perm_new. 2: eassumption. eassumption.
 unfold Z.succ; unfold block; omega.
 (* padding freeable *)
 unfold padding_freeable. 
 left. apply alloc_dummyblocks_perm. eapply Mem.perm_alloc_2. eassumption. 
  assumption.
 eleft. eassumption.   reflexivity. eassumption.
 (* tag main *)
 simpl. rewrite alloc_dummyblocks_tget. eapply Mem.tget_alloc_same. eassumption. eapply Mem.valid_new_block. eassumption.
 (* dummy blocks *)
 reflexivity.
 (* cenv *)
 reflexivity.
 (* env *)
 rewrite VARS. replace (Z.succ (Mem.nextblock m0 - 1)) with (Mem.nextblock m0) by (unfold block; omega). reflexivity.
 (* norepet *)
 rewrite <- map_fst_commut. simpl. rewrite <- VARS. assumption.
 (* variable tags *)
 simpl.
 intros until 1.
 generalize (Mem.alloc_result _ _ _ _ _ _ H2). intro; subst.
 generalize (Mem.inject_nextblock _ _ _ HINJ).
 intro.
 rewrite H9.
 generalize (Mem.nextblock_alloc _ _ _ _ _ _ HALLOC_DUMMY).
 intro.
 assert (Mem.nextblock m = Mem.nextblock m0 - 1) by (unfold block in *; omega).
 rewrite H11.
 erewrite tget_region_alloc_same.
 eauto.
 assumption.
 (* region nextblock (hi bound) *)
 rewrite nextblock_region_alloc.
 reflexivity.
 (* callstack tail *)
 eapply match_callstack_incr_bound.
 eapply match_callstack_invariant.
 eassumption.
 assumption.
 intros. eapply Mem.perm_alloc_4. eassumption.
 eapply region_alloc_perm_old_recip. eassumption.
 eapply Mem.valid_block_alloc. eassumption. assumption.
 generalize (Mem.nextblock_alloc _ _ _ _ _ _ HALLOC_DUMMY).
 unfold block in *. omega.
 intros. eapply alloc_dummyblocks_perm. eapply Mem.perm_alloc_1. eassumption. assumption.
 intros. apply H7.  generalize (Mem.nextblock_alloc _ _ _ _ _ _ HALLOC_DUMMY). unfold block in *. omega.
 intros. rewrite tget_region_alloc_other. eapply Mem.tget_alloc_other. eassumption. generalize (Mem.nextblock_alloc _ _ _ _ _ _ HALLOC_DUMMY). unfold block in *; omega. eapply Mem.valid_block_alloc. eassumption. assumption.
 intros. rewrite alloc_dummyblocks_tget. eapply Mem.tget_alloc_other. eassumption. generalize (Mem.alloc_result _ _ _ _ _ _ H2). unfold block in *; omega. eapply Mem.valid_block_alloc. eassumption. assumption.
 intros.
 rewrite <- H7.
 assumption.
 intro.
 assert (0 <= b - (Mem.nextblock m0 - 1 + 1) < Z.of_nat (length ( l'))) by (unfold block; omega).
 destruct (Z_of_nat_le_inv _ _ H11).
 destruct H12.
 assert (b = Mem.nextblock m0 -  1 + 1 + Z.of_nat x3) by (unfold block; omega).
 subst.
 destruct (nth_error_ex _ _ H13).
 destruct x4. destruct p. destruct p.
 generalize (H6 _ _ _ _ _ H14).
 rewrite H8.
 injection 1; intros; subst.
 generalize (Mem.fresh_block_alloc _ _ _ _ _ _ H2).
 tauto.
 generalize (Mem.nextblock_alloc _ _ _ _ _ _ HALLOC_DUMMY). unfold block; omega.
 generalize (Mem.fresh_block_alloc _ _ _ _ _ _ H2). unfold Mem.valid_block, block. omega.
Qed.

(** * Properties of compile-time approximations of values *)

Definition val_match_approx (a: approx) (v: val) : Prop :=
  match a with
  | Int1 => v = Val.zero_ext 1 v
  | Int7 => v = Val.zero_ext 8 v /\ v = Val.sign_ext 8 v
  | Int8u => v = Val.zero_ext 8 v
  | Int8s => v = Val.sign_ext 8 v
  | Int15 => v = Val.zero_ext 16 v /\ v = Val.sign_ext 16 v
  | Int16u => v = Val.zero_ext 16 v
  | Int16s => v = Val.sign_ext 16 v
  | Float32 => v = Val.singleoffloat v
  | Any => True
  end.

Remark undef_match_approx: forall a, val_match_approx a Vundef.
Proof.
  destruct a; simpl; auto.
Qed.

Lemma val_match_approx_increasing:
  forall a1 a2 v,
  Approx.bge a1 a2 = true -> val_match_approx a2 v -> val_match_approx a1 v.
Proof.
  assert (A: forall v, v = Val.zero_ext 8 v -> v = Val.zero_ext 16 v).
    intros. rewrite H.
    destruct v; simpl; auto. decEq. symmetry. 
    apply Int.zero_ext_widen. omega. 
  assert (B: forall v, v = Val.sign_ext 8 v -> v = Val.sign_ext 16 v).
    intros. rewrite H.
    destruct v; simpl; auto. decEq. symmetry. 
    apply Int.sign_ext_widen. omega.
  assert (C: forall v, v = Val.zero_ext 8 v -> v = Val.sign_ext 16 v).
    intros. rewrite H.
    destruct v; simpl; auto. decEq. symmetry. 
    apply Int.sign_zero_ext_widen. omega.
  assert (D: forall v, v = Val.zero_ext 1 v -> v = Val.zero_ext 8 v).
    intros. rewrite H.
    destruct v; simpl; auto. decEq. symmetry. 
    apply Int.zero_ext_widen. omega. 
  assert (E: forall v, v = Val.zero_ext 1 v -> v = Val.sign_ext 8 v).
    intros. rewrite H.
    destruct v; simpl; auto. decEq. symmetry. 
    apply Int.sign_zero_ext_widen. omega.
  intros. 
  unfold Approx.bge in H; destruct a1; try discriminate; destruct a2; simpl in *; try discriminate; intuition.
Qed.

Lemma approx_of_int_sound:
  forall n, val_match_approx (Approx.of_int n) (Vint n).
Proof.
  unfold Approx.of_int; intros.
  destruct (Int.eq_dec n Int.zero); simpl. subst; auto. 
  destruct (Int.eq_dec n Int.one); simpl. subst; auto. 
  destruct (Int.eq_dec n (Int.zero_ext 7 n)). simpl.
    split.
    decEq. rewrite e. symmetry. apply Int.zero_ext_widen. omega.
    decEq. rewrite e. symmetry. apply Int.sign_zero_ext_widen. omega.
  destruct (Int.eq_dec n (Int.zero_ext 8 n)). simpl; congruence.
  destruct (Int.eq_dec n (Int.sign_ext 8 n)). simpl; congruence.
  destruct (Int.eq_dec n (Int.zero_ext 15 n)). simpl.
    split.
    decEq. rewrite e. symmetry. apply Int.zero_ext_widen. omega. 
    decEq. rewrite e. symmetry. apply Int.sign_zero_ext_widen. omega. 
  destruct (Int.eq_dec n (Int.zero_ext 16 n)). simpl; congruence.
  destruct (Int.eq_dec n (Int.sign_ext 16 n)). simpl; congruence.
  exact I.
Qed.

Lemma approx_of_float_sound:
  forall f, val_match_approx (Approx.of_float f) (Vfloat f).
Proof.
  unfold Approx.of_float; intros. 
  destruct (Float.eq_dec f (Float.singleoffloat f)); simpl; auto. congruence.
Qed.

Lemma approx_of_chunk_sound:
  forall chunk m b ofs v,
  Mem.load chunk m b ofs = Some v ->
  val_match_approx (Approx.of_chunk chunk) v.
Proof.
  intros. exploit Mem.load_cast; eauto. 
  destruct chunk; intros; simpl; auto.
Qed.

Lemma approx_of_unop_sound:
  forall op v1 v a1,
  eval_unop op v1 = Some v ->
  val_match_approx a1 v1 ->
  val_match_approx (Approx.unop op a1) v.
Proof.
  destruct op; simpl; intros; auto; inv H.
  destruct v1; simpl; auto. rewrite Int.zero_ext_idem; auto. omega. 
  destruct v1; simpl; auto. rewrite Int.sign_ext_idem; auto. omega. 
  destruct v1; simpl; auto. rewrite Int.zero_ext_idem; auto. omega. 
  destruct v1; simpl; auto. rewrite Int.sign_ext_idem; auto. omega. 
  destruct v1; simpl; auto. rewrite Float.singleoffloat_idem; auto.
Qed.

Lemma approx_bitwise_and_sound:
  forall a1 v1 a2 v2,
  val_match_approx a1 v1 -> val_match_approx a2 v2 ->
  val_match_approx (Approx.bitwise_and a1 a2) (Val.and v1 v2).
Proof.
  assert (X: forall v1 v2 N, 0 < N < Z_of_nat Int.wordsize ->
            v2 = Val.zero_ext N v2 ->
            Val.and v1 v2 = Val.zero_ext N (Val.and v1 v2)).
    intros. rewrite Val.zero_ext_and in *; auto. 
    rewrite Val.and_assoc. congruence. 
  assert (Y: forall v1 v2 N, 0 < N < Z_of_nat Int.wordsize ->
            v1 = Val.zero_ext N v1 ->
            Val.and v1 v2 = Val.zero_ext N (Val.and v1 v2)).
    intros. rewrite (Val.and_commut v1 v2). apply X; auto.
  assert (P: forall a v, val_match_approx a v -> Approx.bge Int8u a = true ->
               v = Val.zero_ext 8 v).
    intros. apply (val_match_approx_increasing Int8u a v); auto.
  assert (Q: forall a v, val_match_approx a v -> Approx.bge Int16u a = true ->
               v = Val.zero_ext 16 v).
    intros. apply (val_match_approx_increasing Int16u a v); auto.
  assert (R: forall a v, val_match_approx a v -> Approx.bge Int1 a = true ->
               v = Val.zero_ext 1 v).
    intros. apply (val_match_approx_increasing Int1 a v); auto.

  intros; unfold Approx.bitwise_and. 
  destruct (Approx.bge Int1 a1) eqn:?. simpl. apply Y; eauto. compute; auto.
  destruct (Approx.bge Int1 a2) eqn:?. simpl. apply X; eauto. compute; auto.
  destruct (Approx.bge Int8u a1) eqn:?. simpl. apply Y; eauto. compute; auto.
  destruct (Approx.bge Int8u a2) eqn:?. simpl. apply X; eauto. compute; auto.
  destruct (Approx.bge Int16u a1) eqn:?. simpl. apply Y; eauto. compute; auto.
  destruct (Approx.bge Int16u a2) eqn:?. simpl. apply X; eauto. compute; auto.
  simpl; auto. 
Qed.

Lemma approx_bitwise_or_sound:
  forall (sem_op: val -> val -> val) a1 v1 a2 v2,
  (forall a b c, sem_op (Val.and a (Vint c)) (Val.and b (Vint c)) =
                 Val.and (sem_op a b) (Vint c)) ->
  val_match_approx a1 v1 -> val_match_approx a2 v2 ->
  val_match_approx (Approx.bitwise_or a1 a2) (sem_op v1 v2).
Proof.
  intros.
  assert (X: forall v v' N, 0 < N < Z_of_nat Int.wordsize ->
            v = Val.zero_ext N v ->
            v' = Val.zero_ext N v' ->
            sem_op v v' = Val.zero_ext N (sem_op v v')).
    intros. rewrite Val.zero_ext_and in *; auto.
    rewrite H3; rewrite H4. rewrite H. rewrite Val.and_assoc.
    simpl. rewrite Int.and_idem. auto.

  unfold Approx.bitwise_or. 

  destruct (Approx.bge Int1 a1 && Approx.bge Int1 a2) eqn:?.
  destruct (andb_prop _ _ Heqb).
  simpl. apply X. compute; auto. 
  apply (val_match_approx_increasing Int1 a1 v1); auto.
  apply (val_match_approx_increasing Int1 a2 v2); auto.

  destruct (Approx.bge Int8u a1 && Approx.bge Int8u a2) eqn:?.
  destruct (andb_prop _ _ Heqb0).
  simpl. apply X. compute; auto. 
  apply (val_match_approx_increasing Int8u a1 v1); auto.
  apply (val_match_approx_increasing Int8u a2 v2); auto.

  destruct (Approx.bge Int16u a1 && Approx.bge Int16u a2) eqn:?.
  destruct (andb_prop _ _ Heqb1).
  simpl. apply X. compute; auto. 
  apply (val_match_approx_increasing Int16u a1 v1); auto.
  apply (val_match_approx_increasing Int16u a2 v2); auto.

  exact I.
Qed.

Lemma approx_of_binop_sound:
  forall op v1 a1 v2 a2 m v,
  eval_binop op v1 v2 m = Some v ->
  val_match_approx a1 v1 -> val_match_approx a2 v2 ->
  val_match_approx (Approx.binop op a1 a2) v.
Proof.
  assert (OB: forall ob, val_match_approx Int1 (Val.of_optbool ob)).
    destruct ob; simpl. destruct b; auto. auto.
  
  destruct op; intros; simpl Approx.binop; simpl in H; try (exact I); inv H.
  apply approx_bitwise_and_sound; auto.
  apply approx_bitwise_or_sound; auto.
    intros. destruct a; destruct b; simpl; auto.
    rewrite (Int.and_commut i c); rewrite (Int.and_commut i0 c). 
    rewrite <- Int.and_or_distrib. rewrite Int.and_commut. auto.
  apply approx_bitwise_or_sound; auto.
    intros. destruct a; destruct b; simpl; auto.
    rewrite (Int.and_commut i c); rewrite (Int.and_commut i0 c). 
    rewrite <- Int.and_xor_distrib. rewrite Int.and_commut. auto.
  apply OB.
  apply OB.
  apply OB.
Qed.

Lemma approx_unop_is_redundant_sound:
  forall op a v,
  Approx.unop_is_redundant op a = true ->
  val_match_approx a v ->
  eval_unop op v = Some v.
Proof.
  unfold Approx.unop_is_redundant; intros; destruct op; try discriminate.
(* cast8unsigned *)
  assert (V: val_match_approx Int8u v) by (eapply val_match_approx_increasing; eauto).
  simpl in *. congruence.
(* cast8signed *)
  assert (V: val_match_approx Int8s v) by (eapply val_match_approx_increasing; eauto).
  simpl in *. congruence.
(* cast16unsigned *)
  assert (V: val_match_approx Int16u v) by (eapply val_match_approx_increasing; eauto).
  simpl in *. congruence.
(* cast16signed *)
  assert (V: val_match_approx Int16s v) by (eapply val_match_approx_increasing; eauto).
  simpl in *. congruence.
(* singleoffloat *)
  assert (V: val_match_approx Float32 v) by (eapply val_match_approx_increasing; eauto).
  simpl in *. congruence.
Qed.

(** * Compatibility of evaluation functions with respect to memory injections. *)

Remark val_inject_val_of_bool:
  forall f b, val_inject f (Val.of_bool b) (Val.of_bool b).
Proof.
  intros; destruct b; constructor.
Qed.

Remark val_inject_val_of_optbool:
  forall f ob, val_inject f (Val.of_optbool ob) (Val.of_optbool ob).
Proof.
  intros; destruct ob; simpl. destruct b; constructor. constructor.
Qed.

Ltac TrivialExists :=
  match goal with
  | [ |- exists y, Some ?x = Some y /\ val_inject _ _ _ ] =>
      exists x; split; [auto | try(econstructor; eauto)]
  | [ |- exists y, _ /\ val_inject _ (Vint ?x) _ ] =>
      exists (Vint x); split; [eauto with evalexpr | constructor]
  | [ |- exists y, _ /\ val_inject _ (Vfloat ?x) _ ] =>
      exists (Vfloat x); split; [eauto with evalexpr | constructor]
  | _ => idtac
  end.

(** Compatibility of [eval_unop] with respect to [val_inject]. *)

Lemma eval_unop_compat:
  forall f op v1 tv1 v,
  eval_unop op v1 = Some v ->
  val_inject f v1 tv1 ->
  exists tv,
     eval_unop op tv1 = Some tv
  /\ val_inject f v tv.
Proof.
  destruct op; simpl; intros.
  inv H; inv H0; simpl; TrivialExists.
  inv H; inv H0; simpl; TrivialExists.
  inv H; inv H0; simpl; TrivialExists.
  inv H; inv H0; simpl; TrivialExists.
  inv H; inv H0; simpl; TrivialExists.
  inv H; inv H0; simpl; TrivialExists.
  inv H; inv H0; simpl; TrivialExists.
  inv H; inv H0; simpl; TrivialExists.
  inv H; inv H0; simpl; TrivialExists.
  inv H0; simpl in H; inv H. simpl. destruct (Float.intoffloat f0); simpl in *; inv H1. TrivialExists.
  inv H0; simpl in H; inv H. simpl. destruct (Float.intuoffloat f0); simpl in *; inv H1. TrivialExists.
  inv H0; simpl in H; inv H. simpl. TrivialExists.
  inv H0; simpl in H; inv H. simpl. TrivialExists.
Qed.

(** Compatibility of [eval_binop] with respect to [val_inject]. *)

Lemma eval_binop_compat:
  forall f op v1 tv1 v2 tv2 v m tm,
  eval_binop op v1 v2 m = Some v ->
  val_inject f v1 tv1 ->
  val_inject f v2 tv2 ->
  Mem.inject f m tm ->
  exists tv,
     eval_binop op tv1 tv2 tm = Some tv
  /\ val_inject f v tv.
Proof.
  destruct op; simpl; intros.
  inv H; inv H0; inv H1; TrivialExists.
    repeat rewrite Int.add_assoc. decEq. apply Int.add_commut.
    repeat rewrite Int.add_assoc. decEq. apply Int.add_commut.
  inv H; inv H0; inv H1; TrivialExists.
    apply Int.sub_add_l.
    simpl. destruct (zeq b1 b0); auto.
    subst b1. rewrite H in H0; inv H0. 
    rewrite zeq_true. rewrite Int.sub_shifted. auto.
  inv H; inv H0; inv H1; TrivialExists.
  inv H0; try discriminate; inv H1; try discriminate. simpl in *. 
    destruct (Int.eq i0 Int.zero
      || Int.eq i (Int.repr Int.min_signed) && Int.eq i0 Int.mone); inv H; TrivialExists.
  inv H0; try discriminate; inv H1; try discriminate. simpl in *. 
    destruct (Int.eq i0 Int.zero); inv H. TrivialExists.
  inv H0; try discriminate; inv H1; try discriminate. simpl in *. 
    destruct (Int.eq i0 Int.zero
      || Int.eq i (Int.repr Int.min_signed) && Int.eq i0 Int.mone); inv H; TrivialExists.
  inv H0; try discriminate; inv H1; try discriminate. simpl in *. 
    destruct (Int.eq i0 Int.zero); inv H. TrivialExists.
  inv H; inv H0; inv H1; TrivialExists.
  inv H; inv H0; inv H1; TrivialExists.
  inv H; inv H0; inv H1; TrivialExists.
  inv H; inv H0; inv H1; TrivialExists. simpl. destruct (Int.ltu i0 Int.iwordsize); auto.
  inv H; inv H0; inv H1; TrivialExists. simpl. destruct (Int.ltu i0 Int.iwordsize); auto.
  inv H; inv H0; inv H1; TrivialExists. simpl. destruct (Int.ltu i0 Int.iwordsize); auto.
  inv H; inv H0; inv H1; TrivialExists.
  inv H; inv H0; inv H1; TrivialExists.
  inv H; inv H0; inv H1; TrivialExists.
  inv H; inv H0; inv H1; TrivialExists.
  inv H; inv H0; inv H1; TrivialExists. apply val_inject_val_of_optbool.
(* cmpu *)
  inv H; inv H0; inv H1; TrivialExists.
    apply val_inject_val_of_optbool.
    apply val_inject_val_of_optbool.
    apply val_inject_val_of_optbool.
Opaque Int.add.
    unfold Val.cmpu. simpl.
    destruct (zeq b1 b0); subst.
    (* same blocks *)
    rewrite H0 in H. inv H. rewrite zeq_true.
    fold (Mem.weak_valid_pointer m b0 (Int.unsigned ofs1)).
    fold (Mem.weak_valid_pointer m b0 (Int.unsigned ofs0)).
    fold (Mem.weak_valid_pointer tm b2 (Int.unsigned (Int.add ofs1 (Int.repr delta)))).
    fold (Mem.weak_valid_pointer tm b2 (Int.unsigned (Int.add ofs0 (Int.repr delta)))).
    destruct (Mem.weak_valid_pointer m b0 (Int.unsigned ofs1)) eqn:?; auto.
    destruct (Mem.weak_valid_pointer m b0 (Int.unsigned ofs0)) eqn:?; auto.
    rewrite (Mem.weak_valid_pointer_inject_val _ _ _ _ _ _ _ H2 Heqb) by eauto.
    rewrite (Mem.weak_valid_pointer_inject_val _ _ _ _ _ _ _ H2 Heqb0) by eauto.
    rewrite Int.translate_cmpu
      by eauto using Mem.weak_valid_pointer_inject_no_overflow.
    apply val_inject_val_of_optbool.
    (* different source blocks *)
    destruct (Mem.valid_pointer m b1 (Int.unsigned ofs1)) eqn:?; auto.
    destruct (Mem.valid_pointer m b0 (Int.unsigned ofs0)) eqn:?; auto.
    destruct (zeq b2 b3).
    fold (Mem.weak_valid_pointer tm b2 (Int.unsigned (Int.add ofs1 (Int.repr delta)))).
    fold (Mem.weak_valid_pointer tm b3 (Int.unsigned (Int.add ofs0 (Int.repr delta0)))).
    rewrite Mem.valid_pointer_implies
      by (eapply (Mem.valid_pointer_inject_val _ _ _ _ _ _ _ H2 Heqb); eauto).
    rewrite Mem.valid_pointer_implies
      by (eapply (Mem.valid_pointer_inject_val _ _ _ _ _ _ _ H2 Heqb0); eauto).
    exploit Mem.different_pointers_inject; eauto. intros [A|A]; [congruence |].
    destruct c; simpl; auto.
    rewrite Int.eq_false. constructor. congruence.
    rewrite Int.eq_false. constructor. congruence.
    rewrite (Mem.valid_pointer_inject_val _ _ _ _ _ _ _ H2 Heqb) by eauto.
    rewrite (Mem.valid_pointer_inject_val _ _ _ _ _ _ _ H2 Heqb0) by eauto.
    apply val_inject_val_of_optbool.
  (* cmpf *)
  inv H; inv H0; inv H1; TrivialExists. apply val_inject_val_of_optbool.
Qed.

(** * Correctness of Cminor construction functions *)

Lemma make_stackaddr_correct:
  forall sp te tm ofs,
  eval_expr tge (Vptr sp Int.zero) te tm
            (make_stackaddr ofs) (Vptr sp (Int.repr ofs)).
Proof.
  intros; unfold make_stackaddr.
  eapply eval_Econst. simpl. decEq. decEq.
  rewrite Int.add_commut. apply Int.add_zero.
Qed.

Lemma make_globaladdr_correct:
  forall sp te tm id b,
  Genv.find_symbol tge id = Some b ->
  eval_expr tge (Vptr sp Int.zero) te tm
            (make_globaladdr id) (Vptr b Int.zero).
Proof.
  intros; unfold make_globaladdr.
  eapply eval_Econst. simpl. rewrite H. auto.
Qed.

(** Correctness of [make_store]. *)

Inductive val_lessdef_upto (m: int): val -> val -> Prop :=
  | val_lessdef_upto_base:
      forall v1 v2, Val.lessdef v1 v2 -> val_lessdef_upto m v1 v2 
  | val_lessdef_upto_int:
      forall n1 n2, Int.and n1 m = Int.and n2 m -> val_lessdef_upto m (Vint n1) (Vint n2).

 Hint Resolve val_lessdef_upto_base. 

Remark val_lessdef_upto_and:
  forall m v1 v2 p,
  val_lessdef_upto m v1 v2 -> Int.and p m = m ->
  val_lessdef_upto m (Val.and v1 (Vint p)) v2.
Proof.
  intros. inversion H; clear H.
   inversion H1. destruct v2; simpl; auto.  
   apply val_lessdef_upto_int. rewrite Int.and_assoc. congruence.
  simpl. auto. 
  simpl. apply val_lessdef_upto_int. rewrite Int.and_assoc. congruence.  
Qed.

Remark val_lessdef_upto_zero_ext:
  forall m v1 v2 p,
  val_lessdef_upto m v1 v2 -> Int.and (Int.repr (two_p p - 1)) m = m -> 0 < p < 32 ->
  val_lessdef_upto m (Val.zero_ext p v1) v2.
Proof.
  intros. inversion H; clear H.
  inversion H2. destruct v2; simpl; auto. 
  apply val_lessdef_upto_int. rewrite Int.zero_ext_and; auto. 
  rewrite Int.and_assoc. rewrite H0. auto.
  omega.
  simpl; auto.
  simpl. apply val_lessdef_upto_int. rewrite Int.zero_ext_and; auto.
  rewrite Int.and_assoc. rewrite H0. auto. omega. 
Qed.

Remark val_lessdef_upto_sign_ext:
  forall m v1 v2 p,
  val_lessdef_upto m v1 v2 -> Int.and (Int.repr (two_p p - 1)) m = m -> 0 < p < 32 ->
  val_lessdef_upto m (Val.sign_ext p v1) v2.
Proof.
  intros.
  assert (A: forall x, Int.and (Int.sign_ext p x) m = Int.and x m).
    intros. transitivity (Int.and (Int.zero_ext p (Int.sign_ext p x)) m).
    rewrite Int.zero_ext_and; auto. rewrite Int.and_assoc. congruence. omega.
    rewrite Int.zero_ext_sign_ext.
    rewrite Int.zero_ext_and; auto. rewrite Int.and_assoc. congruence. omega. omega.
  inversion H; clear H.
  inversion H2. destruct v2; simpl; auto. 
  apply val_lessdef_upto_int. auto. 
  simpl; auto. 
  simpl. apply val_lessdef_upto_int. rewrite A. auto. 
Qed.

Remark val_lessdef_upto_shru:
  forall m v1 v2 p,
  val_lessdef_upto (Int.shl m p) v1 v2 -> Int.shru (Int.shl m p) p = m ->
  val_lessdef_upto m (Val.shru v1 (Vint p)) (Val.shru v2 (Vint p)).
Proof.
  intros. inversion H; clear H.
  inversion H1; simpl; auto. 
  simpl. destruct (Int.ltu p Int.iwordsize); auto. apply val_lessdef_upto_int.
  rewrite <- H0. repeat rewrite Int.and_shru. congruence.
Qed.

Remark val_lessdef_upto_shr:
  forall m v1 v2 p,
  val_lessdef_upto (Int.shl m p) v1 v2 -> Int.shru (Int.shl m p) p = m ->
  val_lessdef_upto m (Val.shr v1 (Vint p)) (Val.shr v2 (Vint p)).
Proof.
  intros. inversion H; clear H.
  inversion H1; simpl; auto.
  simpl. destruct (Int.ltu p Int.iwordsize); auto. apply val_lessdef_upto_int.
  repeat rewrite Int.shr_and_shru_and; auto.
  rewrite <- H0. repeat rewrite Int.and_shru. congruence.
Qed.

Lemma eval_uncast_int:
  forall m sp te tm a x,
  eval_expr tge sp te tm a x ->
  exists v, eval_expr tge sp te tm (uncast_int m a) v /\ val_lessdef_upto m x v.
Proof.
  assert (EQ: forall p q, Int.eq p q = true -> p = q).
    intros. generalize (Int.eq_spec p q). rewrite H; auto.   
  intros until a. functional induction (uncast_int m a); intros.
  (* cast8unsigned *)
  inv H. simpl in H4; inv H4. exploit IHe; eauto. intros [v [A B]].
  exists v; split; auto. apply val_lessdef_upto_zero_ext; auto. 
  compute; auto.
  exists x; auto. 
  (* cast8signed *)
  inv H. simpl in H4; inv H4. exploit IHe; eauto. intros [v [A B]].
  exists v; split; auto. apply val_lessdef_upto_sign_ext; auto. 
  compute; auto.
  exists x; auto. 
  (* cast16unsigned *)
  inv H. simpl in H4; inv H4. exploit IHe; eauto. intros [v [A B]].
  exists v; split; auto. apply val_lessdef_upto_zero_ext; auto. 
  compute; auto.
  exists x; auto. 
  (* cast16signed *)
  inv H. simpl in H4; inv H4. exploit IHe; eauto. intros [v [A B]].
  exists v; split; auto. apply val_lessdef_upto_sign_ext; auto. 
  compute; auto.
  exists x; auto. 
  (* and *)
  inv H. simpl in H6; inv H6. inv H5. simpl in H0. inv H0.
  exploit IHe; eauto. intros [v [A B]].
  exists v; split; auto. apply val_lessdef_upto_and; auto.
  exists x; auto.
  (* shru *)
  inv H. simpl in H6; inv H6. inv H5. simpl in H0. inv H0.
  exploit IHe; eauto. intros [v [A B]].
  exists (Val.shru v (Vint n)); split.
  econstructor. eauto. econstructor. simpl; reflexivity. auto. 
  apply val_lessdef_upto_shru; auto.
  exists x; auto.
  (* shr *)
  inv H. simpl in H6; inv H6. inv H5. simpl in H0. inv H0.
  exploit IHe; eauto. intros [v [A B]].
  exists (Val.shr v (Vint n)); split.
  econstructor. eauto. econstructor. simpl; reflexivity. auto. 
  apply val_lessdef_upto_shr; auto.
  exists x; auto.
  (* default *)
  exists x; split; auto.
Qed.

Inductive val_lessdef_upto_single: val -> val -> Prop :=
  | val_lessdef_upto_single_base:
      forall v1 v2, Val.lessdef v1 v2 -> val_lessdef_upto_single v1 v2
  | val_lessdef_upto_single_float:
      forall n1 n2, Float.singleoffloat n1 = Float.singleoffloat n2 -> val_lessdef_upto_single (Vfloat n1) (Vfloat n2).

Hint Resolve val_lessdef_upto_single_base.

Lemma eval_uncast_float32:
  forall sp te tm a x,
  eval_expr tge sp te tm a x ->
  exists v, eval_expr tge sp te tm (uncast_float32 a) v /\ val_lessdef_upto_single x v.
Proof.
  intros until a. functional induction (uncast_float32 a); intros.
  (* singleoffloat *)
  inv H. simpl in H4; inv H4. exploit IHe; eauto. intros [v [A B]].
  exists v; split; auto.
  inv B. inv H. destruct v; simpl; auto. 
  apply val_lessdef_upto_single_float. apply Float.singleoffloat_idem. 
  simpl; auto.
  apply val_lessdef_upto_single_float. rewrite H. apply Float.singleoffloat_idem.
  (* default *)
  exists x; auto.
Qed.

Inductive val_content_inject (f: meminj): memory_chunk -> val -> val -> Prop :=
  | val_content_inject_8_signed:
      forall n1 n2, Int.sign_ext 8 n1 = Int.sign_ext 8 n2 ->
      val_content_inject f Mint8signed (Vint n1) (Vint n2)
  | val_content_inject_8_unsigned:
      forall n1 n2, Int.zero_ext 8 n1 = Int.zero_ext 8 n2 ->
      val_content_inject f Mint8unsigned (Vint n1) (Vint n2)
  | val_content_inject_16_signed:
      forall n1 n2, Int.sign_ext 16 n1 = Int.sign_ext 16 n2 ->
      val_content_inject f Mint16signed (Vint n1) (Vint n2)
  | val_content_inject_16_unsigned:
      forall n1 n2, Int.zero_ext 16 n1 = Int.zero_ext 16 n2 ->
      val_content_inject f Mint16unsigned (Vint n1) (Vint n2)
  | val_content_inject_single:
      forall n1 n2, Float.singleoffloat n1 = Float.singleoffloat n2 ->
      val_content_inject f Mfloat32 (Vfloat n1) (Vfloat n2)
  | val_content_inject_base:
      forall chunk v1 v2, val_inject f v1 v2 ->
      val_content_inject f chunk v1 v2.

Hint Resolve val_content_inject_base.

Lemma eval_store_arg:
  forall f sp te tm a v va chunk,
  eval_expr tge sp te tm a va ->
  val_inject f v va ->
  exists vb,
     eval_expr tge sp te tm (store_arg chunk a) vb
  /\ val_content_inject f chunk v vb.
Proof.
  intros.
  assert (DFL: forall v', Val.lessdef va v' -> val_content_inject f chunk v v').
    intros. apply val_content_inject_base. inv H1. auto. inv H0. auto.
  destruct chunk; simpl.
  (* int8signed *)
  exploit (eval_uncast_int (Int.repr 255)); eauto. intros [v' [A B]].
  exists v'; split; auto.
  inv B; auto. inv H0; auto. constructor.
  apply Int.sign_ext_equal_if_zero_equal; auto. omega.
  repeat rewrite Int.zero_ext_and; auto. omega. omega.
  (* int8unsigned *)
  exploit (eval_uncast_int (Int.repr 255)); eauto. intros [v' [A B]].
  exists v'; split; auto.
  inv B; auto. inv H0; auto. constructor.
  repeat rewrite Int.zero_ext_and; auto. omega. omega.
  (* int16signed *)
  exploit (eval_uncast_int (Int.repr 65535)); eauto. intros [v' [A B]].
  exists v'; split; auto.
  inv B; auto. inv H0; auto. constructor.
  apply Int.sign_ext_equal_if_zero_equal; auto. omega.
  repeat rewrite Int.zero_ext_and; auto. omega. omega.
  (* int16unsigned *)
  exploit (eval_uncast_int (Int.repr 65535)); eauto. intros [v' [A B]].
  exists v'; split; auto.
  inv B; auto. inv H0; auto. constructor.
  repeat rewrite Int.zero_ext_and; auto. omega. omega.
  (* int32 *)
  exists va; auto.
  (* float32 *)
  exploit eval_uncast_float32; eauto. intros [v' [A B]].
  exists v'; split; auto.
  inv B; auto. inv H0; auto. constructor. auto.
  (* float64 *)
  exists va; auto.
  (* float64al32 *)
  exists va; auto.
Qed.

Lemma storev_mapped_content_inject:
  forall f chunk m1 b1 o1 v1 n1 m2 a2 v2,
  Mem.inject f m1 m2 ->
  Mem.storev chunk m1 (Vptr b1 o1) v1 = Some n1 ->
  Mem.val_strong_inject f (Vptr b1 o1) a2 ->
  val_content_inject f chunk v1 v2 ->
  forall NVUNDEF: v1 <> Vundef,
  forall ALIGN: (align_chunk chunk | Mem.option_tag_align (Mem.tget m1 b1)),
  exists n2,
    Mem.storev chunk m2 a2 v2 = Some n2 /\ Mem.inject f n1 n2.
Proof.
  intros.
  assert (forall v1',
    (forall b ofs, Mem.store chunk m1 b ofs v1 = Mem.store chunk m1 b ofs v1') ->
    Mem.storev chunk m1 (Vptr b1 o1) v1' = Some n1).
  intros. rewrite <- H0. eauto. 
  inv H2; eapply Mem.storev_mapped_inject;
  try eapply H; try eapply H1; try apply H3; try apply ALIGN; intros.
  rewrite <- Mem.store_int8_sign_ext. rewrite H4. apply Mem.store_int8_sign_ext.
  split; auto.
  rewrite <- Mem.store_int8_zero_ext. rewrite H4. apply Mem.store_int8_zero_ext.
  split; auto.
  rewrite <- Mem.store_int16_sign_ext. rewrite H4. apply Mem.store_int16_sign_ext.
  split; auto.
  rewrite <- Mem.store_int16_zero_ext. rewrite H4. apply Mem.store_int16_zero_ext.
  split; auto.
  rewrite <- Mem.store_float32_truncate. rewrite H4. apply Mem.store_float32_truncate.
  split; auto.
  eauto.
  split; auto.
  intro; contradiction.
Qed.

Lemma make_store_correct:
  forall f sp te tm bl of addr tvaddr rhs tvrhs chunk m vrhs m' fn k,
  eval_expr tge sp te tm addr tvaddr ->
  eval_expr tge sp te tm rhs tvrhs ->
  Mem.storev chunk m (Vptr bl of) vrhs = Some m' ->
  Mem.inject f m tm ->
  val_inject f (Vptr bl of) tvaddr ->
  val_inject f vrhs tvrhs ->
  forall NVUNDEF: vrhs <> Vundef,
    forall ALIGN: (align_chunk chunk | Mem.option_tag_align (Mem.tget m bl)),
  exists tm', exists tvrhs',
  step tge (State fn (make_store chunk addr rhs) k sp te tm)
        E0 (State fn Sskip k sp te tm')
  /\ Mem.storev chunk tm tvaddr tvrhs' = Some tm'
  /\ Mem.inject f m' tm'.
Proof.
  intros. unfold make_store.
  exploit eval_store_arg. eexact H0. eauto. 
  intros [tv [EVAL VCINJ]].
  exploit storev_mapped_content_inject; eauto.
  split; eauto.
  discriminate.
  intros [tm' [STORE MEMINJ]].
  exists tm'; exists tv.
  split. eapply step_store; eauto.
  inversion VCINJ; try congruence. inversion H5; congruence. 
  auto.
Qed.

(** Correctness of [make_unop]. *)

Lemma eval_make_unop:
  forall sp te tm a v op v',
  eval_expr tge sp te tm a v ->
  eval_unop op v = Some v' ->
  exists v'', eval_expr tge sp te tm (make_unop op a) v'' /\ Val.lessdef v' v''.
Proof.
  intros; unfold make_unop. 
  assert (DFL: exists v'', eval_expr tge sp te tm (Eunop op a) v'' /\ Val.lessdef v' v'').
    exists v'; split. econstructor; eauto. auto.
  destruct op; auto; simpl in H0; inv H0.
(* cast8unsigned *)
  exploit (eval_uncast_int (Int.repr 255)); eauto. intros [v1 [A B]].
  exists (Val.zero_ext 8 v1); split. econstructor; eauto. 
  inv B. apply Val.zero_ext_lessdef; auto. simpl.
  repeat rewrite Int.zero_ext_and; auto.
  change (two_p 8 - 1) with 255. rewrite H0. auto.
  omega. omega.
(* cast8signed *)
  exploit (eval_uncast_int (Int.repr 255)); eauto. intros [v1 [A B]].
  exists (Val.sign_ext 8 v1); split. econstructor; eauto. 
  inv B. apply Val.sign_ext_lessdef; auto. simpl.
  replace (Int.sign_ext 8 n2) with (Int.sign_ext 8 n1). auto.
  apply Int.sign_ext_equal_if_zero_equal; auto. omega.
  repeat rewrite Int.zero_ext_and; auto. omega. omega.
(* cast16unsigned *)
  exploit (eval_uncast_int (Int.repr 65535)); eauto. intros [v1 [A B]].
  exists (Val.zero_ext 16 v1); split. econstructor; eauto. 
  inv B. apply Val.zero_ext_lessdef; auto. simpl.
  repeat rewrite Int.zero_ext_and; auto.
  change (two_p 16 - 1) with 65535. rewrite H0. auto.
  omega. omega.
(* cast16signed *)
  exploit (eval_uncast_int (Int.repr 65535)); eauto. intros [v1 [A B]].
  exists (Val.sign_ext 16 v1); split. econstructor; eauto. 
  inv B. apply Val.sign_ext_lessdef; auto. simpl.
  replace (Int.sign_ext 16 n2) with (Int.sign_ext 16 n1). auto.
  apply Int.sign_ext_equal_if_zero_equal; auto. omega.
  repeat rewrite Int.zero_ext_and; auto. omega. omega.
(* singleoffloat *)
  exploit eval_uncast_float32; eauto. intros [v1 [A B]].
  exists (Val.singleoffloat v1); split. econstructor; eauto. 
  inv B. apply Val.singleoffloat_lessdef; auto. simpl. rewrite H0; auto.
Qed.

Lemma make_unop_correct:
  forall f sp te tm a v op v' tv,
  eval_expr tge sp te tm a tv ->
  eval_unop op v = Some v' ->
  val_inject f v tv ->
  exists tv', eval_expr tge sp te tm (make_unop op a) tv' /\ val_inject f v' tv'.
Proof.
  intros. exploit eval_unop_compat; eauto. intros [tv' [A B]].
  exploit eval_make_unop; eauto. intros [tv'' [C D]].
  exists tv''; split; auto. 
  inv D. auto. inv B. auto.
Qed.

(** Correctness of the variable accessor [var_addr] *)

Lemma var_addr_correct:
  forall cenv id f tf e le te sp lo hi m cs tm b,
  match_callstack f m tm (Frame cenv tf e le te sp lo hi :: cs) (Mem.nextblock m) (Mem.nextblock tm) ->
  eval_var_addr ge e id b ->
  exists tv,
     eval_expr tge (Vptr sp Int.zero) te tm (var_addr cenv id) tv
  /\ val_inject f (Vptr b Int.zero) tv.
Proof.
  unfold var_addr; intros.
  assert (match_var f sp e!id cenv!id).
    inv H. inv MENV. auto.
  inv H1; inv H0; try congruence.
  (* local *)
  exists (Vptr sp (Int.repr ofs)); split.
  eapply make_stackaddr_correct.
  congruence.
  (* global *)
  exploit match_callstack_match_globalenvs; eauto. intros [bnd MG]. inv MG.
  exists (Vptr b Int.zero); split.
  eapply make_globaladdr_correct; eauto. rewrite symbols_preserved; auto.
  econstructor; eauto.
Qed.

(** * Semantic preservation for the translation *)

(** The proof of semantic preservation uses simulation diagrams of the
  following form:
<<
       e, m1, s ----------------- sp, te1, tm1, ts
          |                                |
         t|                                |t
          v                                v
       e, m2, out --------------- sp, te2, tm2, tout
>>
  where [ts] is the Cminor statement obtained by translating the
  C#minor statement [s].  The left vertical arrow is an execution
  of a C#minor statement.  The right vertical arrow is an execution
  of a Cminor statement.  The precondition (top vertical bar)
  includes a [mem_inject] relation between the memory states [m1] and [tm1],
  and a [match_callstack] relation for any callstack having
  [e], [te1], [sp] as top frame.  The postcondition (bottom vertical bar)
  is the existence of a memory injection [f2] that extends the injection
  [f1] we started with, preserves the [match_callstack] relation for
  the transformed callstack at the final state, and validates a
  [outcome_inject] relation between the outcomes [out] and [tout].
*)

(** ** Semantic preservation for expressions *)

Remark bool_of_val_inject:
  forall f v tv b,
  Val.bool_of_val v b -> val_inject f v tv -> Val.bool_of_val tv b.
Proof.
  intros. inv H0; inv H; constructor; auto.
Qed.

Lemma transl_constant_correct:
  forall f sp cst v,
  Csharpminor.eval_constant cst = Some v ->
  let (tcst, a) := transl_constant cst in
  exists tv,
     eval_constant tge sp tcst = Some tv
  /\ val_inject f v tv
  /\ val_match_approx a v.
Proof.
  destruct cst; simpl; intros; inv H. 
  exists (Vint i); intuition. apply approx_of_int_sound.
  exists (Vfloat f0); intuition. apply approx_of_float_sound.
Qed.

Lemma transl_expr_correct:
  forall f m tm cenv tf e le te sp lo hi cs
    (MINJ: Mem.inject f m tm)
    (MATCH: match_callstack f m tm
             (Frame cenv tf e le te sp lo hi :: cs)
             (Mem.nextblock m) (Mem.nextblock tm)),
  forall a v,
  Csharpminor.eval_expr ge e le m a v ->
  forall ta app
    (TR: transl_expr cenv a = OK (ta, app)),
  exists tv,
     eval_expr tge (Vptr sp Int.zero) te tm ta tv
  /\ val_inject f v tv
  /\ val_match_approx app v.
Proof.
  induction 3; intros; simpl in TR; try (monadInv TR).
  (* Etempvar *)
  inv MATCH. exploit MTMP; eauto. intros [tv [A B]]. 
  exists tv; split. constructor; auto. split. auto. exact I.
  (* Eaddrof *)
  exploit var_addr_correct; eauto. intros [tv [A B]].
  exists tv; split. auto. split. auto. red. auto.
  (* Econst *)
  exploit transl_constant_correct; eauto. 
  destruct (transl_constant cst) as [tcst a]; inv TR.
  intros [tv [A [B C]]].
  exists tv; split. constructor; eauto. eauto.
  (* Eunop *)
  exploit IHeval_expr; eauto. intros [tv1 [EVAL1 [INJ1 APP1]]].
  unfold Csharpminor.eval_unop in H0. 
  destruct (Approx.unop_is_redundant op x0) eqn:?; inv EQ0.
  (* -- eliminated *)
  exploit approx_unop_is_redundant_sound; eauto. intros. 
  replace v with v1 by congruence.
  exists tv1; auto.
  (* -- preserved *)
  exploit make_unop_correct; eauto. intros [tv [A B]].
  exists tv; split. auto. split. auto. eapply approx_of_unop_sound; eauto.
  (* Ebinop *)
  exploit IHeval_expr1; eauto. intros [tv1 [EVAL1 [INJ1 APP1]]].
  exploit IHeval_expr2; eauto. intros [tv2 [EVAL2 [INJ2 APP2]]].
  exploit eval_binop_compat; eauto. intros [tv [EVAL INJ]].
  exists tv; split. econstructor; eauto. split. auto. eapply approx_of_binop_sound; eauto.
  (* Eload *)
  exploit IHeval_expr; eauto. intros [tv1 [EVAL1 [INJ1 APP1]]].
  exploit Mem.loadv_inject; eauto. intros [tv [LOAD INJ]].
  exists tv; split. econstructor; eauto. inversion INJ. inv H1; congruence. split. inv INJ; auto.
  eapply approx_of_chunk_sound; eauto.
Qed.

Lemma transl_exprlist_correct:
  forall f m tm cenv tf e le te sp lo hi cs
    (MINJ: Mem.inject f m tm)
    (MATCH: match_callstack f m tm
             (Frame cenv tf e le te sp lo hi :: cs)
             (Mem.nextblock m) (Mem.nextblock tm)),
  forall a v,
  Csharpminor.eval_exprlist ge e le m a v ->
  forall ta
    (TR: transl_exprlist cenv a = OK ta),
  exists tv,
     eval_exprlist tge (Vptr sp Int.zero) te tm ta tv
  /\ val_list_inject f v tv.
Proof.
  induction 3; intros; monadInv TR.
  exists (@nil val); split. constructor. constructor.
  exploit transl_expr_correct; eauto. intros [tv1 [EVAL1 [VINJ1 APP1]]].
  exploit IHeval_exprlist; eauto. intros [tv2 [EVAL2 VINJ2]].
  exists (tv1 :: tv2); split. constructor; auto. inv VINJ1; congruence. constructor; auto.
Qed.

(** ** Semantic preservation for statements and functions *)

Inductive match_cont: Csharpminor.cont -> Cminor.cont -> compilenv -> exit_env -> callstack -> Prop :=
  | match_Kstop: forall cenv xenv,
      match_cont Csharpminor.Kstop Kstop cenv xenv nil
  | match_Kseq: forall s k ts tk cenv xenv cs,
      transl_stmt cenv xenv s = OK ts ->
      match_cont k tk cenv xenv cs ->
      match_cont (Csharpminor.Kseq s k) (Kseq ts tk) cenv xenv cs
  | match_Kseq2: forall s1 s2 k ts1 tk cenv xenv cs,
      transl_stmt cenv xenv s1 = OK ts1 ->
      match_cont (Csharpminor.Kseq s2 k) tk cenv xenv cs ->
      match_cont (Csharpminor.Kseq (Csharpminor.Sseq s1 s2) k)
                 (Kseq ts1 tk) cenv xenv cs
  | match_Kblock: forall k tk cenv xenv cs,
      match_cont k tk cenv xenv cs ->
      match_cont (Csharpminor.Kblock k) (Kblock tk) cenv (true :: xenv) cs
  | match_Kblock2: forall k tk cenv xenv cs,
      match_cont k tk cenv xenv cs ->
      match_cont k (Kblock tk) cenv (false :: xenv) cs
  | match_Kcall: forall optid fn e le k tfn sp te tk cenv l' xenv lo hi cs cenv',
      transl_function_aux fn = OK ((cenv, l'), tfn) ->
      match_cont k tk cenv xenv cs ->
      match_cont (Csharpminor.Kcall sp optid fn e le k)
                 (Kcall optid tfn (Vptr sp Int.zero) te tk)
                 cenv' nil
                 (Frame cenv tfn e le te sp lo hi :: cs).

Inductive match_states: Csharpminor.state -> Cminor.state -> Prop :=
  | match_state:
      forall fn s k e le m tfn ts tk sp te tm cenv l' xenv f lo hi cs 
      (TRF: transl_function_aux fn = OK ((cenv, l'), tfn))
      (TR: transl_stmt cenv xenv s = OK ts)
      (MINJ: Mem.inject f m tm)
      (MCS: match_callstack f m tm
               (Frame cenv tfn e le te sp lo hi :: cs)
               (Mem.nextblock m) (Mem.nextblock tm))
      (MK: match_cont k tk cenv xenv cs),
      match_states (Csharpminor.State sp fn s k e le m)
                   (State tfn ts tk (Vptr sp Int.zero) te tm)
  | match_state_seq:
      forall fn s1 s2 k e le m tfn ts1 tk sp te tm cenv l' xenv f lo hi cs 
      (TRF: transl_function_aux fn = OK ((cenv, l'), tfn))
      (TR: transl_stmt cenv xenv s1 = OK ts1)
      (MINJ: Mem.inject f m tm)
      (MCS: match_callstack f m tm
               (Frame cenv tfn e le te sp lo hi :: cs)
               (Mem.nextblock m) (Mem.nextblock tm))
      (MK: match_cont (Csharpminor.Kseq s2 k) tk cenv xenv cs),
      match_states (Csharpminor.State sp fn (Csharpminor.Sseq s1 s2) k e le m)
                   (State tfn ts1 tk (Vptr sp Int.zero) te tm)
  | match_callstate:
      forall symb fd args k m tfd targs tk tm f cs cenv
      (SYMB: Genv.find_funct_ptr ge symb = Some fd)
      (TR: transl_fundef fd = OK tfd)
      (MINJ: Mem.inject f m tm)
      (MCS: match_callstack f m tm cs (Mem.nextblock m) (Mem.nextblock tm))
      (MK: match_cont k tk cenv nil cs)
      (ISCC: Csharpminor.is_call_cont k)
      (ARGSINJ: val_list_inject f args targs),
      match_states (Csharpminor.Callstate fd args k m)
                   (Callstate tfd targs tk tm)
  | match_returnstate:
      forall v k m tv tk tm f cs cenv
      (MINJ: Mem.inject f m tm)
      (MCS: match_callstack f m tm cs (Mem.nextblock m) (Mem.nextblock tm))
      (MK: match_cont k tk cenv nil cs)
      (RESINJ: val_inject f v tv),
      match_states (Csharpminor.Returnstate v k m)
                   (Returnstate tv tk tm).

Remark val_inject_function_pointer:
  forall bound v fd f tv,
  Genv.find_funct ge v = Some fd ->
  match_globalenvs f bound ->
  val_inject f v tv ->
  tv = v.
Proof.
  intros. exploit Genv.find_funct_inv; eauto. intros [b EQ]. subst v.
  rewrite Genv.find_funct_find_funct_ptr in H.
  assert (f b = Some(b, 0)). inv H0. apply DOMAIN. eapply FUNCTIONS; eauto.
  inv H1. rewrite H2 in H5; inv H5. reflexivity.
Qed.

Lemma match_call_cont:
  forall k tk cenv xenv cs,
  match_cont k tk cenv xenv cs ->
  match_cont (Csharpminor.call_cont k) (call_cont tk) cenv nil cs.
Proof.
  induction 1; simpl; auto; econstructor; eauto.
Qed.

Lemma match_is_call_cont:
  forall tfn te sp tm k tk cenv xenv cs,
  match_cont k tk cenv xenv cs ->
  Csharpminor.is_call_cont k ->
  exists tk',
    star step tge (State tfn Sskip tk sp te tm)
               E0 (State tfn Sskip tk' sp te tm)
    /\ is_call_cont tk'
    /\ match_cont k tk' cenv nil cs.
Proof.
  induction 1; simpl; intros; try contradiction.
  econstructor; split. apply star_refl. split. exact I. econstructor; eauto.
  exploit IHmatch_cont; eauto.
  intros [tk' [A B]]. exists tk'; split.
  eapply star_left; eauto. constructor. traceEq. auto.
  econstructor; split. apply star_refl. split. exact I. econstructor; eauto.
Qed.

(** Properties of [switch] compilation *)

Remark switch_table_shift:
  forall n sl base dfl,
  switch_target n (S dfl) (switch_table sl (S base)) =
  S (switch_target n dfl (switch_table sl base)).
Proof.
  induction sl; intros; simpl. auto. destruct (Int.eq n i); auto. 
Qed.

Remark length_switch_table:
  forall sl base1 base2,
  length (switch_table sl base1) = length (switch_table sl base2).
Proof.
  induction sl; intros; simpl. auto. decEq; auto.
Qed.

Inductive transl_lblstmt_cont(cenv: compilenv) (xenv: exit_env): lbl_stmt -> cont -> cont -> Prop :=
  | tlsc_default: forall s k ts,
      transl_stmt cenv (switch_env (LSdefault s) xenv) s = OK ts ->
      transl_lblstmt_cont cenv xenv (LSdefault s) k (Kblock (Kseq ts k))
  | tlsc_case: forall i s ls k ts k',
      transl_stmt cenv (switch_env (LScase i s ls) xenv) s = OK ts ->
      transl_lblstmt_cont cenv xenv ls k k' ->
      transl_lblstmt_cont cenv xenv (LScase i s ls) k (Kblock (Kseq ts k')).

Lemma switch_descent:
  forall cenv xenv k ls body s,
  transl_lblstmt cenv (switch_env ls xenv) ls body = OK s ->
  exists k',
  transl_lblstmt_cont cenv xenv ls k k'
  /\ (forall f sp e m,
      plus step tge (State f s k sp e m) E0 (State f body k' sp e m)).
Proof.
  induction ls; intros. 
  monadInv H. econstructor; split.
  econstructor; eauto.
  intros. eapply plus_left. constructor. apply star_one. constructor. traceEq.
  monadInv H. exploit IHls; eauto. intros [k' [A B]]. 
  econstructor; split.
  econstructor; eauto.
  intros. eapply plus_star_trans. eauto. 
  eapply star_left. constructor. apply star_one. constructor. 
  reflexivity. traceEq.
Qed.

Lemma switch_ascent:
  forall f n sp e m cenv xenv k ls k1,
  let tbl := switch_table ls O in
  let ls' := select_switch n ls in
  transl_lblstmt_cont cenv xenv ls k k1 ->
  exists k2,
  star step tge (State f (Sexit (switch_target n (length tbl) tbl)) k1 sp e m)
             E0 (State f (Sexit O) k2 sp e m)
  /\ transl_lblstmt_cont cenv xenv ls' k k2.
Proof.
  induction ls; intros; unfold tbl, ls'; simpl.
  inv H. econstructor; split. apply star_refl. econstructor; eauto.
  simpl in H. inv H. 
  rewrite Int.eq_sym. destruct (Int.eq i n).
  econstructor; split. apply star_refl. econstructor; eauto. 
  exploit IHls; eauto. intros [k2 [A B]].
  rewrite (length_switch_table ls 1%nat 0%nat). 
  rewrite switch_table_shift.
  econstructor; split. 
  eapply star_left. constructor. eapply star_left. constructor. eexact A. 
  reflexivity. traceEq.
  exact B.
Qed.

Lemma switch_match_cont:
  forall cenv xenv k cs tk ls tk',
  match_cont k tk cenv xenv cs ->
  transl_lblstmt_cont cenv xenv ls tk tk' ->
  match_cont (Csharpminor.Kseq (seq_of_lbl_stmt ls) k) tk' cenv (false :: switch_env ls xenv) cs.
Proof.
  induction ls; intros; simpl.
  inv H0. apply match_Kblock2. econstructor; eauto.
  inv H0. apply match_Kblock2. eapply match_Kseq2. auto. eauto. 
Qed.

Lemma transl_lblstmt_suffix:
  forall n cenv xenv ls body ts,
  transl_lblstmt cenv (switch_env ls xenv) ls body = OK ts ->
  let ls' := select_switch n ls in
  exists body', exists ts',
  transl_lblstmt cenv (switch_env ls' xenv) ls' body' = OK ts'.
Proof.
  induction ls; simpl; intros. 
  monadInv H.
  exists body; econstructor. rewrite EQ; eauto. simpl. reflexivity.
  monadInv H.
  destruct (Int.eq i n).
  exists body; econstructor. simpl. rewrite EQ; simpl. rewrite EQ0; simpl. reflexivity.
  eauto.
Qed.

Lemma switch_match_states:
  forall fn k e le m tfn ts tk sp te tm cenv l' xenv f lo hi cs ls body tk'
    (TRF: transl_function_aux fn = OK ((cenv, l'), tfn))
    (TR: transl_lblstmt cenv (switch_env ls xenv) ls body = OK ts)
    (MINJ: Mem.inject f m tm)
    (MCS: match_callstack f m tm
               (Frame cenv tfn e le te sp lo hi :: cs)
               (Mem.nextblock m) (Mem.nextblock tm))
    (MK: match_cont k tk cenv xenv cs)
    (TK: transl_lblstmt_cont cenv xenv ls tk tk'),
  exists S,
  plus step tge (State tfn (Sexit O) tk' (Vptr sp Int.zero) te tm) E0 S
  /\ match_states (Csharpminor.State sp fn (seq_of_lbl_stmt ls) k e le m) S.
Proof.
  intros. destruct ls; simpl.
  inv TK. econstructor; split. 
    eapply plus_left. constructor. apply star_one. constructor. traceEq.
    eapply match_state; eauto. 
  inv TK. econstructor; split.
    eapply plus_left. constructor. apply star_one. constructor. traceEq.
    eapply match_state_seq; eauto.
    simpl. eapply switch_match_cont; eauto.
Qed.

(** Commutation between [find_label] and compilation *)

Section FIND_LABEL.

Variable lbl: label.
Variable cenv: compilenv.
Variable cs: callstack.

Lemma transl_lblstmt_find_label_context:
  forall xenv ls body ts tk1 tk2 ts' tk',
  transl_lblstmt cenv (switch_env ls xenv) ls body = OK ts ->
  transl_lblstmt_cont cenv xenv ls tk1 tk2 ->
  find_label lbl body tk2 = Some (ts', tk') ->
  find_label lbl ts tk1 = Some (ts', tk').
Proof.
  induction ls; intros.
  monadInv H. inv H0. simpl. simpl in H2. replace x with ts by congruence. rewrite H1. auto.
  monadInv H. inv H0.
  eapply IHls. eauto. eauto. simpl in H6. replace x with ts0 by congruence. simpl.
  rewrite H1. auto.
Qed.

Lemma transl_find_label:
  forall s k xenv ts tk,
  transl_stmt cenv xenv s = OK ts ->
  match_cont k tk cenv xenv cs ->
  match Csharpminor.find_label lbl s k with
  | None => find_label lbl ts tk = None
  | Some(s', k') =>
      exists ts', exists tk', exists xenv',
        find_label lbl ts tk = Some(ts', tk')
     /\ transl_stmt cenv xenv' s' = OK ts'
     /\ match_cont k' tk' cenv xenv' cs
  end

with transl_lblstmt_find_label:
  forall ls xenv body k ts tk tk1,
  transl_lblstmt cenv (switch_env ls xenv) ls body = OK ts ->
  match_cont k tk cenv xenv cs ->
  transl_lblstmt_cont cenv xenv ls tk tk1 ->
  find_label lbl body tk1 = None ->
  match Csharpminor.find_label_ls lbl ls k with
  | None => find_label lbl ts tk = None
  | Some(s', k') =>
      exists ts', exists tk', exists xenv',
        find_label lbl ts tk = Some(ts', tk')
     /\ transl_stmt cenv xenv' s' = OK ts'
     /\ match_cont k' tk' cenv xenv' cs
  end.
Proof.
  intros. destruct s; try (monadInv H); simpl; auto.
  (* seq *)
  exploit (transl_find_label s1). eauto. eapply match_Kseq. eexact EQ1. eauto.  
  destruct (Csharpminor.find_label lbl s1 (Csharpminor.Kseq s2 k)) as [[s' k'] | ].
  intros [ts' [tk' [xenv' [A [B C]]]]]. 
  exists ts'; exists tk'; exists xenv'. intuition. rewrite A; auto.
  intro. rewrite H. apply transl_find_label with xenv; auto. 
  (* ifthenelse *)
  exploit (transl_find_label s1). eauto. eauto.  
  destruct (Csharpminor.find_label lbl s1 k) as [[s' k'] | ].
  intros [ts' [tk' [xenv' [A [B C]]]]]. 
  exists ts'; exists tk'; exists xenv'. intuition. rewrite A; auto.
  intro. rewrite H. apply transl_find_label with xenv; auto. 
  (* loop *)
  apply transl_find_label with xenv. auto. econstructor; eauto. simpl. rewrite EQ; auto.
  (* block *)
  apply transl_find_label with (true :: xenv). auto. constructor; auto. 
  (* switch *)
  exploit switch_descent; eauto. intros [k' [A B]].
  eapply transl_lblstmt_find_label. eauto. eauto. eauto. reflexivity.  
  (* return *)
  destruct o; monadInv H; auto.
  (* label *)
  destruct (ident_eq lbl l).
  exists x; exists tk; exists xenv; auto.
  apply transl_find_label with xenv; auto.

  intros. destruct ls; monadInv H; simpl.
  (* default *)
  inv H1. simpl in H3. replace x with ts by congruence. rewrite H2.
  eapply transl_find_label; eauto.
  (* case *)
  inv H1. simpl in H7.
  exploit (transl_find_label s). eauto. eapply switch_match_cont; eauto. 
  destruct (Csharpminor.find_label lbl s (Csharpminor.Kseq (seq_of_lbl_stmt ls) k)) as [[s' k''] | ].
  intros [ts' [tk' [xenv' [A [B C]]]]].
  exists ts'; exists tk'; exists xenv'; intuition. 
  eapply transl_lblstmt_find_label_context; eauto. 
  simpl. replace x with ts0 by congruence. rewrite H2. auto.
  intro.
  eapply transl_lblstmt_find_label. eauto. auto. eauto. 
  simpl. replace x with ts0 by congruence. rewrite H2. auto. 
Qed.

End FIND_LABEL.

Lemma transl_find_label_body:
  forall cenv l' xenv f tf k tk cs lbl s' k',
  transl_function_aux f = OK ((cenv, l'), tf) ->
  match_cont k tk cenv xenv cs ->
  Csharpminor.find_label lbl f.(Csharpminor.fn_body) (Csharpminor.call_cont k) = Some (s', k') ->
  exists ts', exists tk', exists xenv',
     find_label lbl tf.(fn_body) (call_cont tk) = Some(ts', tk')
  /\ transl_stmt cenv xenv' s' = OK ts'
  /\ match_cont k' tk' cenv xenv' cs.
Proof.
  intros. unfold transl_function_aux in H. destruct (zle (stack_size (Csharpminor.fn_vars f)) Int.max_unsigned); try discriminate. monadInv H. simpl. 
  exploit transl_find_label. eexact EQ1. eapply match_call_cont. eexact H0.
  instantiate (1 := lbl). rewrite H1. auto.
Qed.

(** The simulation diagram. *)

Fixpoint seq_left_depth (s: Csharpminor.stmt) : nat :=
  match s with
  | Csharpminor.Sseq s1 s2 => S (seq_left_depth s1)
  | _ => O
  end.

Definition measure (S: Csharpminor.state) : nat :=
  match S with
  | Csharpminor.State sp fn s k e le m => seq_left_depth s
  | _ => O
  end.

Ltac var v e :=
  pose (v := e);
  let H := fresh in
  assert (H: v = e) by reflexivity;
    try rewrite <- H;
    revert H;
    generalize v; clear v.


Lemma volatile_load_inject:
  forall F V (ge: Genv.t F V) f chunk m b ofs t v b' ofs' m',
  meminj_preserves_globals ge f ->
  volatile_load ge chunk m b ofs t v ->
  val_inject f (Vptr b ofs) (Vptr b' ofs') ->
  Mem.inject f m m' ->
  exists v', volatile_load ge chunk m' b' ofs' t v' /\ val_inject f v v'.
Proof.
  intros. inv H0.
  inv H1. exploit (proj1 H); eauto. intros EQ; rewrite H8 in EQ; inv EQ.
  rewrite Int.add_zero. exists (Val.load_result chunk v0); split.
  constructor; auto. 
  apply val_load_result_inject. eapply eventval_match_inject_2; eauto.
  exploit Mem.loadv_inject; eauto. simpl; eauto. simpl; intros [v' [A B]]. exists v'; split; auto.
  constructor; auto. rewrite <- H3. inv H1. eapply meminj_preserves_block_is_volatile; eauto.
  erewrite <- Mem.tget_inject; eauto.
  eapply Zdivide_trans. eassumption. inv H1. eapply Mem.tget_inject_align_block. eassumption. eassumption.
  inversion B. inv H0; congruence.
  destruct B; assumption.
Qed.

Lemma volatile_store_inject:
  forall F V (ge: Genv.t F V) f chunk m1 b ofs v t m2 m1' b' ofs' v',
  meminj_preserves_globals ge f ->
  volatile_store ge chunk m1 b ofs v t m2 ->
  val_inject f (Vptr b ofs) (Vptr b' ofs') ->
  val_inject f v v' ->
  Mem.inject f m1 m1' ->
  exists m2',
       volatile_store ge chunk m1' b' ofs' v' t m2'
    /\ Mem.inject f m2 m2'
    /\ mem_unchanged_on (loc_unmapped f) m1 m2
    /\ mem_unchanged_on (loc_out_of_reach f m1) m1' m2'.
Proof.
  intros. inv H0.
  inv H1. exploit (proj1 H); eauto. intros EQ; rewrite H9 in EQ; inv EQ.
  rewrite Int.add_zero. exists m1'.
  split. constructor; auto.  eapply eventval_match_inject; eauto.
  split. auto. split. red; auto. red; auto. 
  assert (Mem.storev chunk m1 (Vptr b ofs) v = Some m2). simpl; auto.
  exploit Mem.storev_mapped_inject; eauto. split. eassumption. intro; contradiction. intros [m2' [A B]].
  inv H1. exists m2'; intuition. 
  constructor; auto. rewrite <- H4. eapply meminj_preserves_block_is_volatile; eauto.
  erewrite <- Mem.tget_inject; eauto. eapply Zdivide_trans. eassumption. eapply Mem.tget_inject_align_block; eauto.
  inversion H2; congruence.
  split; intros. eapply Mem.perm_store_1; eauto.
  rewrite <- H6. eapply Mem.load_store_other; eauto. 
  left. exploit (H1 ofs0). generalize (size_chunk_pos chunk0). omega. 
  unfold loc_unmapped. congruence.
  split; intros. eapply Mem.perm_store_1; eauto.
  rewrite <- H6. eapply Mem.load_store_other; eauto.
  destruct (eq_block b0 b'); auto. subst b0; right.
  assert (EQ: Int.unsigned (Int.add ofs (Int.repr delta)) = Int.unsigned ofs + delta).
    eapply Mem.address_inject; eauto with mem.
  unfold Mem.storev in A. rewrite EQ in A. rewrite EQ.
  apply (Intv.range_disjoint' (ofs0, ofs0 + size_chunk chunk0)
                              (Int.unsigned ofs + delta, Int.unsigned ofs + delta + size_chunk chunk)).
  red; intros; red; intros. exploit (H1 x H7). eauto.
  exploit Mem.store_valid_access_3. eexact H0. intros [C D].
  apply Mem.perm_cur_max. apply Mem.perm_implies with Writable; auto with mem.
  apply C. red in H8; simpl in H8. omega. 
  auto.
  simpl. generalize (size_chunk_pos chunk0). omega.
  simpl. generalize (size_chunk_pos chunk). omega.
Qed.

Section EXTERNAL_CALL_INJECT.

Let ec_inject (external_call_extc: extcall_sem) := forall F V (ge: Genv.t F V) vargs m1 t vres m2 f m1' vargs',
    meminj_preserves_globals ge f ->
    external_call_extc _ _ ge vargs m1 t vres m2 ->
    Mem.inject f m1 m1' ->
    val_list_inject f vargs vargs' ->
    exists f', exists vres', exists m2',
       external_call_extc _ _ ge vargs' m1' t vres' m2'
    /\ val_inject f' vres vres'
    /\ Mem.inject f' m2 m2'
    /\ mem_unchanged_on (loc_unmapped f) m1 m2
    /\ mem_unchanged_on (loc_out_of_reach f m1) m1' m2'
    /\ inject_incr f f'
    /\ inject_separated f f' m1 m1'.

Lemma ec_inject_io: forall name sg, ec_inject (extcall_io_sem name sg).
Proof.
  unfold ec_inject. intros. inv H0.
  exists f; exists vres; exists m1'; intuition try congruence.
  econstructor; eauto.
  eapply eventval_list_match_inject; eauto.
  eapply eventval_match_inject_2; eauto.
  red; auto.
  red; auto.
Qed.

Lemma ec_inject_load: forall chunk, ec_inject (volatile_load_sem chunk).
Proof.
  unfold ec_inject; intros.
  inv H0. inv H2. inv H7. inversion H5; subst. 
  exploit volatile_load_inject; eauto. intros [v' [A B]]. 
  exists f; exists v'; exists m1'; intuition try congruence. constructor; auto.
  red; auto. red; auto. 
Qed.

Lemma ec_inject_store: forall chunk, ec_inject (volatile_store_sem chunk).
Proof.
  unfold ec_inject; intros.
  inv H0. inv H2. inv H7. inv H8. inversion H5; subst.
  exploit volatile_store_inject; eauto. intros [m2' [A [B [C D]]]].
  exists f; exists Vundef; exists m2'; intuition try congruence. constructor; auto. constructor. 
Qed.

Lemma ec_inject_load_global: forall chunk id ofs, ec_inject (volatile_load_global_sem chunk id ofs).
Proof.
  unfold ec_inject; intros.
  inv H0. inv H2.
  assert (val_inject f (Vptr b ofs) (Vptr b ofs)). 
    exploit (proj1 H); eauto. intros EQ. econstructor. eauto. rewrite Int.add_zero; auto.
  exploit volatile_load_inject; eauto. intros [v' [A B]].
  exists f; exists v'; exists m1'; intuition try congruence. econstructor; eauto. 
  red; auto. red; auto. 
Qed.

Lemma ec_inject_store_global: forall chunk id ofs, ec_inject (volatile_store_global_sem chunk id ofs).
Proof.
  unfold ec_inject; intros.
  rewrite volatile_store_global_charact in H0. destruct H0 as [b [P Q]]. 
  exploit (proj1 H). eauto. intros EQ. 
  assert (val_inject f (Vptr b ofs) (Vptr b ofs)). econstructor; eauto. rewrite Int.add_zero; auto.
  generalize (ec_inject_store). unfold ec_inject. intro K. exploit K. eauto. eexact Q. eauto. eauto. 
  intros [f' [vres' [m2' [A [B [C [D [E G]]]]]]]].
  exists f'; exists vres'; exists m2'. split; auto.
  rewrite volatile_store_global_charact. exists b; auto. 
Qed.

Lemma ec_inject_malloc: ec_inject extcall_malloc_sem.
Proof.
  assert (UNCHANGED:
    forall (P: block -> Z -> Prop) m n m' b m'',
    Mem.alloc m (-4) (Int.unsigned n) (Mem.Tag_heap Mem.Heap_freestore) = (m', b) ->
    Mem.store Mint32 m' b (-4) (Vint n) = Some m'' ->
    mem_unchanged_on P m m'').
  intros; split; intros.
  eauto with mem.
  transitivity (Mem.load chunk m' b0 ofs). 
  eapply Mem.load_store_other; eauto. left. 
  apply Mem.valid_not_valid_diff with m; eauto with mem.
  eapply Mem.load_alloc_other; eauto. 

  unfold ec_inject; intros.
  inv H0. inv H2. inv H6. inv H8.
  exploit Mem.alloc_parallel_inject; eauto. apply Zle_refl. apply Zle_refl. 
  congruence.
  intros [f' [m3' [b' [ALLOC [A [B [C D]]]]]]].
  exploit Mem.store_mapped_inject. eexact A. eauto. eauto. 
  instantiate (1 := Vint n). split; auto.
  erewrite Mem.tget_alloc_same. 2: eassumption. exists 2. reflexivity.
  intros [m2' [E G]].
  Opaque inject_separated.
  exists f'; exists (Vptr b' Int.zero); exists m2'; intuition.
  econstructor; eauto.
  econstructor. eauto. auto.
  eapply UNCHANGED; eauto.
  eapply UNCHANGED; eauto.
  Transparent inject_separated. red; intros. destruct (eq_block b1 b). 
  subst b1. rewrite C in H2. inv H2. eauto with mem. 
  rewrite D in H2. congruence. auto. 
Qed. 

Lemma ec_inject_free: ec_inject extcall_free_sem.
Proof.
  assert (UNCHANGED:
    forall (P: block -> Z -> Prop) m b lo hi m',
    Mem.free m b lo hi = Some m' ->
    lo < hi ->
    (forall b' ofs, P b' ofs -> b' <> b \/ ofs < lo \/ hi <= ofs) ->
    mem_unchanged_on P m m').
  intros; split; intros.
  eapply Mem.perm_free_1; eauto.
  rewrite <- H3. eapply Mem.load_free; eauto. 
  destruct (eq_block b0 b); auto. right. right. 
  apply (Intv.range_disjoint' (ofs, ofs + size_chunk chunk) (lo, hi)).
  red; intros. apply Intv.notin_range. simpl. exploit H1; eauto. intuition. 
  simpl; generalize (size_chunk_pos chunk); omega.
  simpl; omega.

  unfold ec_inject; intros.
  inv H0. inv H2. inv H7. inv H9.
  exploit Mem.load_inject; eauto. 
  rewrite TAG. exists 2. reflexivity.
  intros [vsz [A B]]. inv B.  inv H0.
  assert (Mem.range_perm m1 b (Int.unsigned lo - 4) (Int.unsigned lo + Int.unsigned sz) Cur Freeable).
    eapply Mem.free_range_perm; eauto.
  exploit Mem.address_inject; eauto. 
    apply Mem.perm_implies with Freeable; auto with mem.
    eapply H0. instantiate (1 := lo). omega. 
  intro EQ.
  exploit Mem.free_parallel_inject.
   eassumption.
   eassumption.
   eassumption.
  destruct 1. destruct H7.
  esplit. esplit. esplit. split.
econstructor.
  rewrite EQ. replace (Int.unsigned lo + delta - 4) with (Int.unsigned lo - 4 + delta) by omega.
  eauto. auto. 
  rewrite EQ. replace (Int.unsigned lo + delta - 4) with (Int.unsigned lo - 4 + delta) by omega. 
  replace (Int.unsigned lo + delta + Int.unsigned sz) with (Int.unsigned lo + Int.unsigned sz + delta) by omega. eassumption.
 erewrite <- Mem.tget_inject. 2: eassumption. eapply Mem.tget_inject_heap_freestore. eassumption. eassumption. assumption.
 split. constructor.
 split. eassumption.
 
split.  eapply UNCHANGED; eauto. omega. intros.  
  red in H9. left. congruence. 

split.  eapply UNCHANGED; eauto. omega. intros.
  destruct (eq_block b' b2); auto. subst b'. right.
  assert (~(Int.unsigned lo + delta - 4 <= ofs < Int.unsigned lo + delta + Int.unsigned sz)).
    red; intros. elim (H9 _ _ H6). 
    apply Mem.perm_cur_max. apply Mem.perm_implies with Freeable; auto with mem.
    apply H0. omega.
  omega.
split. auto. unfold inject_separated. intros; congruence.
Qed.
  
Lemma ec_inject_memcpy: forall sz al, ec_inject (extcall_memcpy_sem sz al).
Proof.
  unfold ec_inject; intros.
  intros. inv H0. inv H2. inv H14. inv H15. inv H11. inv H12.
  exploit Mem.loadbytes_length; eauto. intros LEN.
  assert (RPSRC: Mem.range_perm m1 bsrc (Int.unsigned osrc) (Int.unsigned osrc + sz) Cur Nonempty).
    eapply Mem.range_perm_implies. eapply Mem.loadbytes_range_perm; eauto. auto with mem.
  assert (RPDST: Mem.range_perm m1 bdst (Int.unsigned odst) (Int.unsigned odst + sz) Cur Nonempty).
    replace sz with (Z_of_nat (length bytes)).
    eapply Mem.range_perm_implies. eapply Mem.storebytes_range_perm; eauto. auto with mem.
    rewrite LEN. apply nat_of_Z_eq. omega.
  assert (PSRC: Mem.perm m1 bsrc (Int.unsigned osrc) Cur Nonempty).
    apply RPSRC. omega.
  assert (PDST: Mem.perm m1 bdst (Int.unsigned odst) Cur Nonempty).
    apply RPDST. omega.
  exploit Mem.address_inject.  eauto. eexact PSRC. eauto. intros EQ1.
  exploit Mem.address_inject.  eauto. eexact PDST. eauto. intros EQ2.
  exploit Mem.loadbytes_inject; eauto. intros [bytes2 [A B]].
  exploit Mem.storebytes_mapped_inject; eauto. intros [m2' [C D]].
  exists f; exists Vundef; exists m2'.
  split. econstructor; try rewrite EQ1; try rewrite EQ2; eauto. 
  eapply Mem.aligned_area_inject with (m := m1); eauto.
  eapply Mem.aligned_area_inject with (m := m1); eauto.
  erewrite <- Mem.tget_inject; eauto. eapply Zdivide_trans. eexact ALIGN_SRC. eapply Mem.tget_inject_align_block. eassumption. eassumption.
  erewrite <- Mem.tget_inject; eauto. eapply Zdivide_trans. eexact ALIGN_DST. eapply Mem.tget_inject_align_block. eassumption. eassumption.
  eapply Mem.disjoint_or_equal_inject with (m := m1); eauto.
  apply Mem.range_perm_max with Cur; auto.
  apply Mem.range_perm_max with Cur; auto.
  split. constructor.
  split. auto.
  split. red; split; intros. eauto with mem. 
  rewrite <- H2. eapply Mem.load_storebytes_other; eauto. 
  destruct (eq_block b bdst); auto. subst b. 
  assert (loc_unmapped f bdst ofs). apply H0. generalize (size_chunk_pos chunk); omega. 
  red in H12. congruence.
  split. red; split; intros. eauto with mem.
  rewrite <- H2. eapply Mem.load_storebytes_other; eauto. 
  destruct (eq_block b b0); auto. subst b0; right.
  rewrite <- (list_forall2_length B). rewrite LEN. rewrite nat_of_Z_eq; try omega.
  apply (Intv.range_disjoint' (ofs, ofs + size_chunk chunk)
                              (Int.unsigned odst + delta0, Int.unsigned odst + delta0 + sz)); simpl.
  red; unfold Intv.In; simpl; intros; red; intros.
  eapply (H0 x H12). eauto. apply Mem.perm_cur_max. apply RPDST. omega. 
  generalize (size_chunk_pos chunk); omega.
  omega.
  split. apply inject_incr_refl.
  red; intros; congruence.
Qed.

Lemma ec_inject_annot: forall text targs,
                         ec_inject (extcall_annot_sem text targs).
Proof.
  unfold ec_inject; intros.
  inv H0.
  exists f; exists Vundef; exists m1'; intuition try congruence.
  econstructor; eauto.
  eapply eventval_list_match_inject; eauto.
  auto.
  auto.
  red; auto.
  red; auto.
Qed. 

Lemma ec_inject_annot_val: forall text targ, 
  ec_inject (extcall_annot_val_sem text targ).
Proof.
  unfold ec_inject; intros.
  inv H0. inv H2. inv H7.
  exists f; exists v'; exists m1'; intuition try congruence.
  econstructor; eauto.
  eapply eventval_match_inject; eauto.
  red; auto.
  red; auto.
Qed.

Lemma external_call_mem_inject:
  forall extc, ec_inject (external_call extc).
Proof.
  destruct extc; simpl; intros.
   apply ec_inject_io.
   apply ec_inject_io.
   apply ec_inject_load.
   apply ec_inject_store.
   apply ec_inject_load_global.
   apply ec_inject_store_global.
   apply ec_inject_malloc.
   apply ec_inject_free.
   apply ec_inject_memcpy.
   apply ec_inject_annot.
   apply ec_inject_annot_val.
   apply ec_inject_annot.
Qed.

End EXTERNAL_CALL_INJECT.

Lemma transl_step_correct:
  forall S1 t S2, Csharpminor.step true ge S1 t S2 ->
  forall T1, match_states S1 T1 ->
  (exists T2, plus step tge T1 t T2 /\ match_states S2 T2)
  \/ (measure S2 < measure S1 /\ t = E0 /\ match_states S2 T1)%nat.
Proof.
  induction 1; intros T1 MSTATE; inv MSTATE.

(* skip seq *)
  monadInv TR. left.
  revert MK.
  var sk (Csharpminor.Kseq s k).
  intros.
  induction MK; try discriminate.
   inv H.
   econstructor; split. 
   apply plus_one. constructor.
   econstructor; eauto.
  inv H.
  econstructor; split.
  apply plus_one. constructor.
  eapply match_state_seq; eauto. 
  exploit IHMK; eauto. intros [T2 [A B]].
  exists T2; split. eapply plus_left. constructor. apply plus_star; eauto. traceEq.
  auto.
(* skip block *)
  monadInv TR. left.
  revert MK.
  var sk (Csharpminor.Kblock k).
  induction 2; try discriminate.
  inv H.
  econstructor; split.
  apply plus_one. constructor.
  econstructor; eauto.
  exploit IHMK; eauto. intros [T2 [A B]].
  exists T2; split. eapply plus_left. constructor. apply plus_star; eauto. traceEq.
  auto.
(* skip call *)
  monadInv TR. left.
  unfold transl_function_aux in TRF. destruct (zle (stack_size (Csharpminor.fn_vars f)) Int.max_unsigned); try discriminate. monadInv TRF. generalize (check_offsets_correct _ _ EQ). intro HEQ. rewrite HEQ in H0.
  rewrite map_snd_commut in H0.
  rewrite <- region_free_region_free_aux in H0.
  assert (lo = Z.succ db).
   inv MCS. inv MENV. unfold Z.succ. omega.
  subst.
  exploit match_is_call_cont; eauto. intros [tk' [A [B C]]]. 
  exploit match_callstack_region_free ; eauto. 
  simpl. eassumption.
  simpl.
  intros [tm' [P [Q R]]].
  econstructor; split.
  eapply plus_right. eexact A. apply step_skip_call. auto. eauto. traceEq.
  econstructor; eauto.

(* set *)
  monadInv TR.
  exploit transl_expr_correct; eauto. intros [tv [EVAL [VINJ APP]]].
  left; econstructor; split.
  apply plus_one. econstructor; eauto. 
  econstructor; eauto. 
  eapply match_callstack_set_temp; eauto. 

(* store *)
  monadInv TR.
  exploit transl_expr_correct. eauto. eauto. eexact H. eauto. 
  intros [tv1 [EVAL1 [VINJ1 APP1]]].
  exploit transl_expr_correct. eauto. eauto. eexact H0. eauto. 
  intros [tv2 [EVAL2 [VINJ2 APP2]]].
  exploit make_store_correct. eexact EVAL1. eexact EVAL2. eauto. eauto. auto. auto. assumption.
  assumption.
  intros [tm' [tv' [EXEC [STORE' MINJ']]]].
  left; econstructor; split.
  apply plus_one. eexact EXEC.
  econstructor; eauto.
  inv VINJ1; simpl in H1; try discriminate. unfold Mem.storev in STORE'.
  rewrite (Mem.nextblock_store _ _ _ _ _ _ H1).
  rewrite (Mem.nextblock_store _ _ _ _ _ _ STORE').
  eapply match_callstack_invariant with f0 m tm; eauto.
  intros. eapply Mem.perm_store_2; eauto. 
  intros. eapply Mem.perm_store_1; eauto.
  intros. eapply Mem.tget_store; eauto.
  intros. eapply Mem.tget_store; eauto.

(* call *)
  simpl in H1. exploit functions_translated; eauto. intros [tfd [FIND TRANS]].
  monadInv TR.
  exploit transl_expr_correct; eauto. intros [tvf [EVAL1 [VINJ1 APP1]]].
  assert (tvf = vf).
    exploit match_callstack_match_globalenvs; eauto. intros [bnd MG].
    eapply val_inject_function_pointer; eauto.
  subst tvf.
  exploit transl_exprlist_correct; eauto.
  intros [tvargs [EVAL2 VINJ2]].
  left; econstructor; split.
  apply plus_one. eapply step_call; eauto.
  apply sig_preserved; eauto.
  destruct vf; try discriminate.
  simpl in H1. destruct (Int.eq_dec i Int.zero); try discriminate.
  econstructor; eauto.
  eapply match_Kcall with (cenv' := cenv); eauto.
  red; auto.

(* builtin *)
  monadInv TR.
  exploit transl_exprlist_correct; eauto.
  intros [tvargs [EVAL2 VINJ2]].
  exploit match_callstack_match_globalenvs; eauto. intros [hi' MG].
  exploit external_call_mem_inject; eauto. 
  eapply inj_preserves_globals; eauto.
  intros [f' [vres' [tm' [EC [VINJ [MINJ' [UNMAPPED [OUTOFREACH [INCR SEPARATED]]]]]]]]].
  left; econstructor; split.
  apply plus_one. econstructor. eauto. 
  eapply external_call_symbols_preserved; eauto.
  exact symbols_preserved. eexact varinfo_preserved.
  assert (MCS': match_callstack f' m' tm'
                 (Frame cenv tfn e le te db lo hi :: cs)
                 (Mem.nextblock m') (Mem.nextblock tm')).
    apply match_callstack_incr_bound with (Mem.nextblock m) (Mem.nextblock tm).
    eapply match_callstack_external_call; eauto.
    intros. eapply external_call_max_perm; eauto.
    intros. case_eq (Mem.tget m b). eapply ec_tget; eauto. eapply external_call_spec. intros. absurd (Mem.tget m b = None). eapply Mem.tget_valid. assumption. assumption.
    intros. case_eq (Mem.tget tm sp). eapply ec_tget; eauto. eapply external_call_spec. intros. absurd (Mem.tget tm sp = None). eapply Mem.tget_valid. assumption. assumption.    
    omega. omega. 
    eapply external_call_nextblock; eauto.
    eapply external_call_nextblock; eauto.
  econstructor; eauto.
Opaque PTree.set.
  unfold set_optvar. destruct optid; simpl. 
  eapply match_callstack_set_temp; eauto. 
  auto.

(* seq *)
  monadInv TR. 
  left; econstructor; split.
  apply plus_one. constructor. 
  econstructor; eauto.
  econstructor; eauto.
(* seq 2 *)
  right. split. auto. split. auto. econstructor; eauto. 

(* ifthenelse *)
  monadInv TR.
  exploit transl_expr_correct; eauto. intros [tv [EVAL [VINJ APP]]].
  left; exists (State tfn (if b then x1 else x2) tk (Vptr db Int.zero) te tm); split.
  apply plus_one. eapply step_ifthenelse; eauto. eapply bool_of_val_inject; eauto.
  econstructor; eauto. destruct b; auto.

(* loop *)
  monadInv TR.
  left; econstructor; split.
  apply plus_one. constructor. 
  econstructor; eauto.
  econstructor; eauto. simpl. rewrite EQ; auto. 

(* block *)
  monadInv TR.
  left; econstructor; split.
  apply plus_one. constructor. 
  econstructor; eauto.
  econstructor; eauto.

(* exit seq *)
  monadInv TR. left.
  revert MK. var v (Csharpminor.Kseq s k). intros. revert s k H. induction MK; try discriminate.
  injection 1; intros; subst.
  econstructor; split. 
  apply plus_one. constructor.
  econstructor; eauto. simpl. auto.
  injection 1; intros; subst.
  exploit IHMK; eauto. intros [T2 [A B]].
  exists T2; split; auto. eapply plus_left. constructor. apply plus_star; eauto. traceEq.
  intros; subst.
  exploit IHMK; eauto. intros [T2 [A B]].
  exists T2; split; auto. eapply plus_left.
  simpl. constructor. apply plus_star; eauto. traceEq.

(* exit block 0 *)
  monadInv TR. left.
  revert MK. var v (Csharpminor.Kblock k). intros. revert k H. induction MK; try discriminate.
  injection 1; intros; subst.
  econstructor; split.
  simpl. apply plus_one. constructor. 
  econstructor; eauto.
  intros; subst.
  exploit IHMK; eauto. intros [T2 [A B]].
  exists T2; split; auto. simpl. 
  eapply plus_left. constructor. apply plus_star; eauto. traceEq.

(* exit block n+1 *)
  monadInv TR. left.
  revert MK. var v (Csharpminor.Kblock k). intros. revert k H. induction MK; try discriminate.
  injection 1; intros; subst.
  econstructor; split.
  simpl. apply plus_one. constructor. 
  econstructor; eauto. auto. 
  intros; subst.
  exploit IHMK; eauto. intros [T2 [A B]].
  exists T2; split; auto. simpl. 
  eapply plus_left. constructor. apply plus_star; eauto. traceEq.

(* switch *)
  monadInv TR. left.
  exploit transl_expr_correct; eauto. intros [tv [EVAL [VINJ APP]]].
  inv VINJ.
  exploit switch_descent; eauto. intros [k1 [A B]].
  exploit switch_ascent; eauto. intros [k2 [C D]].
  exploit transl_lblstmt_suffix; eauto. simpl. intros [body' [ts' E]].
  exploit switch_match_states; eauto. intros [T2 [F G]].
  exists T2; split.
  eapply plus_star_trans. eapply B.  
  eapply star_left. econstructor; eauto. 
  eapply star_trans. eexact C. 
  apply plus_star. eexact F.
  reflexivity. reflexivity. traceEq.
  auto.

(* return none *)
  monadInv TR. left.
  unfold transl_function_aux in TRF. destruct (zle (stack_size (Csharpminor.fn_vars f)) Int.max_unsigned); try discriminate. monadInv TRF. generalize (check_offsets_correct _ _ EQ). intro HEQ. rewrite HEQ in H.
  rewrite map_snd_commut in H.
  rewrite <- region_free_region_free_aux in H.
  assert (lo = Z.succ db).
   inv MCS. inv MENV. unfold Z.succ. omega.
  subst.
  exploit match_callstack_region_free; eauto.
   simpl. eassumption.
  simpl.
  intros [tm' [A [B C]]].
  econstructor; split.
  apply plus_one. eapply step_return_0. eauto.
  econstructor; eauto. eapply match_call_cont; eauto.

(* return some *)
  monadInv TR. left. 
  exploit transl_expr_correct; eauto. intros [tv [EVAL [VINJ APP]]].
  unfold transl_function_aux in TRF. destruct (zle (stack_size (Csharpminor.fn_vars f)) Int.max_unsigned); try discriminate. monadInv TRF. generalize (check_offsets_correct _ _ EQ0). intro HEQ. rewrite HEQ in H0.
  rewrite map_snd_commut in H0.
  rewrite <- region_free_region_free_aux in H0.
  assert (lo = Z.succ db).
   inv MCS. inv MENV. unfold Z.succ. omega.
  subst.
  exploit match_callstack_region_free; eauto.
   eassumption.
  simpl; intros [tm' [A [B C]]].
  econstructor; split.
  apply plus_one. eapply step_return_1. eauto. eauto. 
  econstructor; eauto. eapply match_call_cont; eauto.

(* label *)
  monadInv TR.
  left; econstructor; split.
  apply plus_one. constructor.
  econstructor; eauto.

(* goto *)
  monadInv TR.
  exploit transl_find_label_body; eauto. 
  intros [ts' [tk' [xenv' [A [B C]]]]].
  left; econstructor; split.
  apply plus_one. apply step_goto. eexact A. 
  econstructor; eauto.

(* internal call *)
  monadInv TR. monadInv EQ. destruct x0.  destruct p. generalize EQ0. unfold transl_function_aux.
  destruct (zle (stack_size (Csharpminor.fn_vars f)) Int.max_unsigned); try congruence.
  intro. monadInv EQ1. revert EQ0.
  set (tf := mkfunction (Csharpminor.fn_id f) (Csharpminor.fn_sig f) 
                        (Csharpminor.fn_params f)
                        (Csharpminor.fn_temps f)
                        (l)
                        (stack_size (Csharpminor.fn_vars f))
                        x0) in *.
  intro.
  exploit match_callstack_function_entry; eauto. 
  simpl.
  intros [f2 [tm2 [sp [ALLOC2 [INJ2 [INCR2 CALLSTACK2]]]]]].
  left; econstructor; split.
  apply plus_one. econstructor; simpl; eauto.
  generalize (Mem.inject_nextblock _ _ _ MINJ).
  intro.
  generalize (Mem.alloc_result _ _ _ _ _ _ DUMMYBLOCK).
  generalize (Mem.alloc_result _ _ _ _ _ _ ALLOC2).
  intros; subst.
  rewrite H6 at 1.
  econstructor. eassumption. eauto. eassumption. rewrite <- H6. eassumption.
  inv MK; simpl in ISCC; contradiction || econstructor; eauto.

(* external call *)
  monadInv TR. 
  exploit match_callstack_match_globalenvs; eauto. intros [hi MG].
  exploit external_call_mem_inject; eauto. 
  eapply inj_preserves_globals; eauto.
  intros [f' [vres' [tm' [EC [VINJ [MINJ' [UNMAPPED [OUTOFREACH [INCR SEPARATED]]]]]]]]].
  left; econstructor; split.
  apply plus_one. econstructor.
  eapply external_call_symbols_preserved; eauto.
  exact symbols_preserved. eexact varinfo_preserved.
  econstructor; eauto.
  apply match_callstack_incr_bound with (Mem.nextblock m) (Mem.nextblock tm).
  eapply match_callstack_external_call; eauto.  
  intros. eapply external_call_max_perm; eauto.
  intros. case_eq (Mem.tget m b).
   intros. eapply ec_tget. eapply external_call_spec. eassumption. assumption.
  intro. apply Mem.tget_valid in H0. contradiction.
  intros. case_eq (Mem.tget tm sp).
   intros. eapply ec_tget. eapply external_call_spec. eassumption. assumption.
  intro. apply Mem.tget_valid in H0. contradiction.
  omega. omega.
  eapply external_call_nextblock; eauto.
  eapply external_call_nextblock; eauto.

(* return *)
  inv MK. simpl.
  left; econstructor; split.
  apply plus_one. econstructor; eauto. 
  unfold set_optvar. destruct optid; simpl; econstructor; eauto.
  eapply match_callstack_set_temp; eauto. 
Qed.

Lemma match_globalenvs_init:
  forall m,
  Genv.init_mem prog = Some m ->
  match_globalenvs (Mem.flat_inj (Mem.nextblock m)) (Mem.nextblock m).
Proof.
  intros. constructor.
  apply Mem.nextblock_pos.
  intros. unfold Mem.flat_inj. apply zlt_true; auto. 
  intros. unfold Mem.flat_inj in H0. 
  destruct (zlt b1 (Mem.nextblock m)); congruence.
  intros. eapply Genv.find_symbol_not_fresh; eauto.
  intros. eapply Genv.find_funct_ptr_not_fresh; eauto.
  intros. eapply Genv.find_var_info_not_fresh; eauto. 
Qed.

Lemma transl_initial_states:
  forall S, Csharpminor.initial_state prog S ->
  exists R, Cminor.initial_state tprog R /\ match_states S R.
Proof.
  induction 1.
  exploit function_ptr_translated; eauto. intros [tf [FIND TR]].
  econstructor; split.
  econstructor.
  apply (Genv.init_mem_transf_partial _ _ TRANSL). eauto. 
  simpl. fold tge. rewrite symbols_preserved.
  replace (prog_main tprog) with (prog_main prog). eexact H0.
  symmetry. unfold transl_program in TRANSL.
  eapply transform_partial_program_main; eauto.
  eexact FIND. 
  rewrite <- H2. apply sig_preserved; auto.
  eapply match_callstate with (f := Mem.flat_inj (Mem.nextblock m0)) (cs := @nil frame) (cenv := PTree.empty Z).
  eassumption.
  assumption.
  eapply Mem.neutral_inject.
  eapply Genv.initmem_inject_neutral; eauto.
  apply mcs_nil with (Mem.nextblock m0). apply match_globalenvs_init; auto. omega. omega.
  constructor. red; auto.
  constructor. 
Qed.

Lemma transl_final_states:
  forall S R r,
  match_states S R -> Csharpminor.final_state S r -> Cminor.final_state R r.
Proof.
  intros. inv H0. inv H. inv MK. inv RESINJ. constructor.
Qed.

Theorem transl_program_correct:
  forward_simulation (Csharpminor.semantics true prog) (Cminor.semantics tprog).
Proof.
  eapply forward_simulation_star; eauto.
  eexact symbols_preserved.
  eexact transl_initial_states.
  eexact transl_final_states.
  eexact transl_step_correct. 
Qed.

End TRANSLATION.

