(* *********************************************************************)
(*                                                                     *)
(*              The Compcert verified compiler                         *)
(*                                                                     *)
(*          Xavier Leroy, INRIA Paris-Rocquencourt                     *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique.  All rights reserved.  This file is distributed       *)
(*  under the terms of the INRIA Non-Commercial License Agreement.     *)
(*                                                                     *)
(* *********************************************************************)

(** Abstract syntax and semantics for the Csharpminor language. *)

Require Import Coqlib.
Require Import Maps.
Require Import AST.
Require Import Integers.
Require Import Floats.
Require Import Values.
Require Import Memory.
Require Import Events.
Require Import Globalenvs.
Require Cminor.
Require Import Smallstep.
Require Import StrongInject.

Section WITHMEM.
Context `{SI: Mem.StrongInject}.

(** Abstract syntax *)

(** Csharpminor is a low-level imperative language structured in expressions,
  statements, functions and programs.  Expressions include
  reading temporary variables, taking the address of a variable,
  constants, arithmetic operations, and dereferencing addresses. *)

Inductive constant : Type :=
  | Ointconst: int -> constant          (**r integer constant *)
  | Ofloatconst: float -> constant.     (**r floating-point constant *)

Definition unary_operation : Type := Cminor.unary_operation.
Definition binary_operation : Type := Cminor.binary_operation.

Inductive expr : Type :=
  | Evar : ident -> expr                (**r reading a temporary variable  *)
  | Eaddrof : ident -> expr             (**r taking the address of a variable *)
  | Econst : constant -> expr           (**r constants *)
  | Eunop : unary_operation -> expr -> expr  (**r unary operation *)
  | Ebinop : binary_operation -> expr -> expr -> expr (**r binary operation *)
  | Eload : memory_chunk -> expr -> expr. (**r memory read *)

(** Statements include expression evaluation, temporary variable assignment,
  memory stores, function calls, an if/then/else conditional,
  infinite loops, blocks and early block exits, and early function returns.
  [Sexit n] terminates prematurely the execution of the [n+1] enclosing
  [Sblock] statements. *)

Definition label := ident.

Inductive stmt : Type :=
  | Sskip: stmt
  | Sset : ident -> expr -> stmt
  | Sstore : memory_chunk -> expr -> expr -> stmt
  | Scall : option ident -> signature -> expr -> list expr -> stmt
  | Sbuiltin : option ident -> external_function -> list expr -> stmt
  | Sseq: stmt -> stmt -> stmt
  | Sifthenelse: expr -> stmt -> stmt -> stmt
  | Sloop: stmt -> stmt
  | Sblock: stmt -> stmt
  | Sexit: nat -> stmt
  | Sswitch: expr -> lbl_stmt -> stmt
  | Sreturn: option expr -> stmt
  | Slabel: label -> stmt -> stmt
  | Sgoto: label -> stmt

with lbl_stmt : Type :=
  | LSdefault: stmt -> lbl_stmt
  | LScase: int -> stmt -> lbl_stmt -> lbl_stmt.

(** Functions are composed of a return type, a list of parameter names,
  a list of local variables with their sizes, a list of temporary variables,
  and a statement representing the function body.  *)

Record function: Type := mkfunction {
  fn_id: ident;
  fn_sig: signature;  
  fn_params: list ident;
  fn_vars: list (ident * (Z * Z * option Z));
  fn_temps: list ident;
  fn_body: stmt
}.

Definition fundef := AST.fundef function.

Definition program : Type := AST.program fundef unit.

Definition funsig (fd: fundef) :=
  match fd with
  | Internal f => fn_sig f
  | External ef => ef_sig ef
  end.

(** * Operational semantics *)

(** Three evaluation environments are involved:
- [genv]: global environments, map symbols and functions to memory blocks,
    and maps symbols to variable informations (type [var_kind])
- [env]: local environments, map local variables 
    to pairs (memory block, size)
- [temp_env]: local environments, map temporary variables to
    their current values.
*)

Definition genv := Genv.t fundef unit.
Definition env := PTree.t (block * Z).
Definition temp_env := PTree.t val.

Definition empty_env : env := PTree.empty (block * Z).
Definition empty_temp_env : temp_env := PTree.empty val.

(** Initialization of temporary variables *)

Fixpoint create_undef_temps (temps: list ident) : temp_env :=
  match temps with
  | nil => PTree.empty val
  | id :: temps' => PTree.set id Vundef (create_undef_temps temps')
 end.

(** Initialization of temporaries that are parameters. *)

Fixpoint bind_parameters (formals: list ident) (args: list val)
                         (le: temp_env) : option temp_env :=
 match formals, args with
 | nil, nil => Some le
 | id :: xl, v :: vl => bind_parameters xl vl (PTree.set id v le)
 | _, _ => None
 end. 

(** Continuations *)

Inductive cont: Type :=
  | Kstop: cont                         (**r stop program execution *)
  | Kseq: stmt -> cont -> cont          (**r execute stmt, then cont *)
  | Kblock: cont -> cont                (**r exit a block, then do cont *)
  | Kcall: block -> option ident -> function -> env -> temp_env -> cont -> cont.
                                        (**r return to caller *)

(** States *)

Inductive state `{M: Mem.MEM mem}: Type :=
  | State:                      (**r Execution within a function *)
      forall (b: block)                 (**r dummy empty block *)
             (f: function)              (**r currently executing function  *)
             (s: stmt)                  (**r statement under consideration *)
             (k: cont)                  (**r its continuation -- what to do next *)
             (e: env)                   (**r current local environment *)
             (le: temp_env)             (**r current temporary environment *)
             (m: mem),                  (**r current memory state *)
      state
  | Callstate:                  (**r Invocation of a function *)
      forall (f: fundef)                (**r function to invoke *)
             (args: list val)           (**r arguments provided by caller *)
             (k: cont)                  (**r what to do next  *)
             (m: mem),                  (**r memory state *)
      state
  | Returnstate:                (**r Return from a function *)
      forall (v: val)                   (**r Return value *)
             (k: cont)                  (**r what to do next *)
             (m: mem),                  (**r memory state *)
      state.

(** Pop continuation until a call or stop *)

Fixpoint call_cont (k: cont) : cont :=
  match k with
  | Kseq s k => call_cont k
  | Kblock k => call_cont k
  | _ => k
  end.

Definition is_call_cont (k: cont) : Prop :=
  match k with
  | Kstop => True
  | Kcall _ _ _ _ _ _ => True
  | _ => False
  end.

(** Resolve [switch] statements. *)

Fixpoint select_switch (n: int) (sl: lbl_stmt) {struct sl} : lbl_stmt :=
  match sl with
  | LSdefault _ => sl
  | LScase c s sl' => if Int.eq c n then sl else select_switch n sl'
  end.

Fixpoint seq_of_lbl_stmt (sl: lbl_stmt) : stmt :=
  match sl with
  | LSdefault s => s
  | LScase c s sl' => Sseq s (seq_of_lbl_stmt sl')
  end.

(** Find the statement and manufacture the continuation 
  corresponding to a label *)

Fixpoint find_label (lbl: label) (s: stmt) (k: cont) 
                    {struct s}: option (stmt * cont) :=
  match s with
  | Sseq s1 s2 =>
      match find_label lbl s1 (Kseq s2 k) with
      | Some sk => Some sk
      | None => find_label lbl s2 k
      end
  | Sifthenelse a s1 s2 =>
      match find_label lbl s1 k with
      | Some sk => Some sk
      | None => find_label lbl s2 k
      end
  | Sloop s1 =>
      find_label lbl s1 (Kseq (Sloop s1) k)
  | Sblock s1 =>
      find_label lbl s1 (Kblock k)
  | Sswitch a sl =>
      find_label_ls lbl sl k
  | Slabel lbl' s' =>
      if ident_eq lbl lbl' then Some(s', k) else find_label lbl s' k
  | _ => None
  end

with find_label_ls (lbl: label) (sl: lbl_stmt) (k: cont) 
                   {struct sl}: option (stmt * cont) :=
  match sl with
  | LSdefault s => find_label lbl s k
  | LScase _ s sl' =>
      match find_label lbl s (Kseq (seq_of_lbl_stmt sl') k) with
      | Some sk => Some sk
      | None => find_label_ls lbl sl' k
      end
  end.

(** Evaluation of operator applications. *)

Definition eval_constant (cst: constant) : option val :=
  match cst with
  | Ointconst n => Some (Vint n)
  | Ofloatconst n => Some (Vfloat n)
  end.

Definition eval_unop := Cminor.eval_unop.

Definition eval_binop := Cminor.eval_binop.

(** Allocation of local variables at function entry.  Each variable is
  bound to the reference to a fresh block of the appropriate size. *)

Section WITH_OFFSETS.
Variable with_offsets: bool.

Function stack_size (vars: list (ident * (Z * Z * option Z))) {struct vars} : Z :=
  match vars with
    | nil => 0
    | (_, (sz, _, _)) :: vars' => sz + stack_size vars'
  end.


Definition call_condition (l: list (ident * (Z * Z * option Z))) : Prop :=
  if with_offsets
  then (
      (forall id sz al of, In (id, (sz, al, of)) l ->
                           exists off, of = Some off /\ (al | off))
      /\
      (forall id1 sz1 al1 of1,
         In (id1, (sz1, al1, Some of1)) l ->
         forall id2 sz2 al2 of2,
           In (id2, (sz2, al2, Some of2)) l ->
           id1 <> id2 ->
           of1 + sz1 <= of2 \/ of2 + sz2 <= of1)
      /\
      (forall id1 sz1 al1 of1,
         In (id1, (sz1, al1, Some of1)) l ->
         ((al1 | of1) /\ 0 <= of1 /\ of1 + sz1 <= stack_size l))
    ) else (
      forall id sz al of, In (id, (sz, al, of)) l -> of = None
    ).

Require Import Errors.

Function check_offsets (accu: list (ident * (Z * Z * Z))) (vars: list (ident * (Z * Z * option Z))): res (list (ident * (Z * Z * Z))) :=
  match vars with
    | nil => OK (List.rev accu)
    | (id, (sz, al, Some of)) :: vars' => 
      check_offsets ((id, (sz, al, of)) :: accu) vars'
    | _ => Error  (msg "You forgot the Csharpminor to Csharpminor2 pass (offset computation)!")
  end.

Lemma check_offsets_correct_aux:
  forall v a r,
    check_offsets a v = OK r ->
    List.map (fun i => match i with (id, (sz, of, al)) => (id, (sz, of, Some al)) end) r =
    List.rev (List.map (fun i => match i with (id, (sz, of, al)) => (id, (sz, of, Some al)) end) a) ++ v.
Proof.
  intros until r.
  functional induction (check_offsets a v); simpl in *; try discriminate.
   injection 1; intros; subst.
   rewrite <- app_nil_end.
   rewrite map_rev. reflexivity.
  intro.
  rewrite (IHr0 H); clear IHr0 H.
  rewrite app_ass.
  reflexivity.
Qed.

Corollary check_offsets_correct:
  forall v r,
    check_offsets nil v = OK r ->
    v = List.map (fun i => match i with (id, (sz, of, al)) => (id, (sz, of, Some al)) end) r.
Proof.
  intros. rewrite (check_offsets_correct_aux _ _ _ H). reflexivity.
Qed.                              

Lemma stack_size_region_size: forall l, stack_size (List.map (fun i => match i with (id, (sz, of, al)) => (id, (sz, of, Some al)) end) l) = Mem.region_size (List.map snd l).
Proof.
  induction l; simpl.
   reflexivity.
  rewrite IHl; clear IHl.
  destruct a. destruct p. destruct p. reflexivity.
Qed.

Function assign_variables_aux (e: env) (b: block) (vars: list (ident * (Z * Z * option Z))) {struct vars}: env :=
  match vars with
    | nil => e
    | (id, (sz, _, _)) :: vars' => assign_variables_aux (PTree.set id (b, sz) e) (Zsucc b) vars'
  end.

Function check_no_offsets (accu: list (ident * (Z * Z))) (vars: list (ident * (Z * Z * option Z))): option (list (ident * (Z * Z))) :=
  match vars with
    | nil => Some (rev accu)
    | (id, (sz, al, None)) :: vars' => check_no_offsets ((id, (sz, al)) :: accu) vars'
    | (_, (_, _, Some _)) :: _ => None
  end.

Lemma check_no_offsets_correct_aux:
  forall l accu l',
    check_no_offsets accu l = Some l' ->
    List.map (fun idsza => match idsza with (id, (sz, al)) => (id, (sz, al, None)) end) (rev accu) ++ l = List.map (fun idsza => match idsza with (id, (sz, al)) => (id, (sz, al, None)) end) l'.
Proof.
  induction l; simpl.
   intros. inv H. rewrite <- app_nil_end. reflexivity.   
  destruct a. destruct p. destruct p. destruct o.
   discriminate.
  intros. 
  symmetry.
  erewrite <- IHl.  2: eassumption.
  simpl. rewrite map_app. simpl. rewrite app_ass. reflexivity.
Qed.

Lemma check_no_offsets_correct:
  forall l l',
    check_no_offsets nil l = Some l' ->
    List.map (fun idsza => match idsza with (id, (sz, al)) => (id, (sz, al, None)) end)  l' = l.
Proof.
  intros. exploit check_no_offsets_correct_aux. eassumption.
  simpl. congruence.
Qed.

Function region_alloc' (b: block) (f: ident) (m: mem) (l: list (ident * (Z * Z))) {struct l} : mem :=
  match l with
    | nil => m
    | (v, (sz, al)) :: l' =>
      match Mem.alloc m 0 sz (Mem.Tag_stack (Mem.Stack_var (Mem.Svar f v b sz al None))) with
        | (m', _) =>
          region_alloc' b f m' l'
      end
  end.

Inductive alloc_variables (m: mem) (b: block) (i: ident) (l: list (ident * (Z * Z * option Z))) (e: env) (m': mem) (osz: option Z) : Prop :=
| alloc_variables_with_offsets
    (WO_TRUE: with_offsets = true)
    l' (Hl': check_offsets nil l = OK l')    
    (Hb: b = Mem.nextblock m - 1)
    (He: e = assign_variables_aux empty_env (Zsucc b) l)
    (Hm': m' = Mem.region_alloc b i m l')
    (Hosz: osz = Some (stack_size l))
| alloc_variables_without_offsets
    (WO_FALSE: with_offsets = false)
    l' (Hcheck: check_no_offsets nil l = Some l')
    (Hb: b = Mem.nextblock m - 1)
    (He: e = assign_variables_aux empty_env (Zsucc b) l)
    (Hm': m' = region_alloc' b i m l')
    (Hosz: osz = None)
.

Function region_free_aux (b: block) (m: mem) (l: list (Z * Z * option Z)) {struct l}: option mem :=
  match l with
    | nil => Some m
    | (sz, _, _) :: l' =>
      match Mem.free m b 0 sz with
        | Some m' => region_free_aux (Zsucc b) m' l'
        | None => None
      end
  end.

Lemma region_free_region_free_aux:
  forall b m l,
    Mem.region_free b m l = region_free_aux b m (List.map (fun sao => match sao with (sz, al, of) => (sz, al, Some of) end) l).
Proof.
  intros.
  functional induction (Mem.region_free b m l); simpl.
   reflexivity.
  rewrite e0. assumption.
  rewrite e0. reflexivity.
Qed.

(*
Inductive alloc_variables (b: block): env -> mem ->
                           list (ident * (Z * Z * option Z)) ->
                           env -> mem -> option Z -> Prop :=
  | alloc_variables_nil:
      forall e m osz
             (HSZ: if with_offsets then osz = None else osz = Some 0),
      alloc_variables b e m nil e m osz
  | alloc_variables_cons:
      forall e m id sz al of otot otot' vars m1 b1 m2 e2,
      Mem.alloc m 0 sz
                (Mem.Tag_stack
                   (Mem.Stack_var b sz al of)) = (m1, b1) ->      
      alloc_variables b (PTree.set id (b1, sz) e) m1 vars e2 m2 otot' ->
      forall SUM: match (with_offsets, of, otot, otot') with
        | (true, Some of', Some tot, Some tot') => tot' = of' + tot
        | (false, None, None, None) => True
        | _ => False
      end,
      alloc_variables b e m ((id, (sz, al, of)) :: vars) e2 m2 otot.
*)

(** List of blocks mentioned in an environment, with low and high bounds 

Definition block_of_binding (id_b_sz: ident * (block * Z)) :=
  match id_b_sz with (id, (b, sz)) => (b, 0, sz) end.

Definition blocks_of_env (e: env) : list (block * Z * Z) :=
  List.map block_of_binding (PTree.elements e).
*)

Section RELSEM.

Variable ge: genv.

(* Evaluation of the address of a variable: 
   [eval_var_addr prg ge e id b] states that variable [id] 
   in environment [e] evaluates to block [b]. *)

Inductive eval_var_addr: env -> ident -> block -> Prop :=
  | eval_var_addr_local:
      forall e id b sz,
      PTree.get id e = Some (b, sz) ->
      eval_var_addr e id b
  | eval_var_addr_global:
      forall e id b,
      PTree.get id e = None ->
      Genv.find_symbol ge id = Some b ->
      eval_var_addr e id b.

(** Evaluation of an expression: [eval_expr prg e m a v] states
  that expression [a], in initial memory state [m] and local
  environment [e], evaluates to value [v]. *)

Section EVAL_EXPR.

Variable e: env.
Variable le: temp_env.
Variable m: mem.

Inductive eval_expr: expr -> val -> Prop :=
  | eval_Evar: forall id v,
      le!id = Some v ->
      eval_expr (Evar id) v
  | eval_Eaddrof: forall id b,
      eval_var_addr e id b ->
      eval_expr (Eaddrof id) (Vptr b Int.zero)
  | eval_Econst: forall cst v,
      eval_constant cst = Some v ->
      eval_expr (Econst cst) v
  | eval_Eunop: forall op a1 v1 v,
      eval_expr a1 v1 ->
      eval_unop op v1 = Some v ->
      eval_expr (Eunop op a1) v
  | eval_Ebinop: forall op a1 a2 v1 v2 v,
      eval_expr a1 v1 ->
      eval_expr a2 v2 ->
      eval_binop op v1 v2 m = Some v ->
      eval_expr (Ebinop op a1 a2) v
  | eval_Eload: forall chunk a b o v,
      eval_expr a (Vptr b o) ->
      Mem.load chunk m b (Int.unsigned o) = Some v ->
        forall (ALIGN: (* forall b' sz al of, 
                         Mem.tget m b = Some (Mem.Tag_stack (Mem.Stack_var b' sz al of)) -> *)
                         (align_chunk chunk | Mem.option_tag_align (Mem.tget m b))),
        forall (NVUNDEF: v <> Vundef),
      eval_expr (Eload chunk a) v.

(** Evaluation of a list of expressions:
  [eval_exprlist prg e m al vl] states that the list [al] of
  expressions evaluate to the list [vl] of values.  The other
  parameters are as in [eval_expr]. *)

Inductive eval_exprlist: list expr -> list val -> Prop :=
  | eval_Enil:
      eval_exprlist nil nil
  | eval_Econs: forall a1 al v1 vl,
      eval_expr a1 v1 -> eval_exprlist al vl ->
      forall (NVUNDEF: v1 <> Vundef),
      eval_exprlist (a1 :: al) (v1 :: vl).

End EVAL_EXPR.


(** One step of execution *)


Inductive step: state -> trace -> state -> Prop :=

  | step_skip_seq: forall db f s k e le m,
      step (State db f Sskip (Kseq s k) e le m)
        E0 (State db f s k e le m)
  | step_skip_block: forall db f k e le m,
      step (State db f Sskip (Kblock k) e le m)
        E0 (State db f Sskip k e le m)
  | step_skip_call: forall db f k e le m m',
      is_call_cont k ->
      region_free_aux (Zsucc db) m (List.map snd f.(fn_vars)) = Some m' ->
      step (State db f Sskip k e le m)
        E0 (Returnstate Vundef k m')

  | step_set: forall db f id a k e le m v,
      eval_expr e le m a v ->
      step (State db f (Sset id a) k e le m)
        E0 (State db f Sskip k e (PTree.set id v le) m)

  | step_store: forall db f chunk addr a k e le m b o v m',
      eval_expr e le m addr (Vptr b o) ->
      eval_expr e le m a v ->
      Mem.store chunk m b (Int.unsigned o) v = Some m' ->
      forall (NVUNDEF: v <> Vundef),
        forall (ALIGN: (* forall b' sz al of, 
                         Mem.tget m b = Some (Mem.Tag_stack (Mem.Stack_var b' sz al of)) -> *)
                         (align_chunk chunk | Mem.option_tag_align (Mem.tget m b))
                         ),
      step (State db f (Sstore chunk addr a) k e le m)
        E0 (State db f Sskip k e le m')

  | step_call: forall db f optid sig a bl k e le m vf vargs fd,
      eval_expr e le m a vf ->
      eval_exprlist e le m bl vargs ->
      Genv.find_funct ge vf = Some fd ->
      funsig fd = sig ->
      step (State db f (Scall optid sig a bl) k e le m)
        E0 (Callstate fd vargs (Kcall db optid f e le k) m)

  | step_builtin: forall db f optid ef bl k e le m vargs t vres m',
      eval_exprlist e le m bl vargs ->
      external_call ef ge vargs m t vres m' ->
      step (State db f (Sbuiltin optid ef bl) k e le m)
         t (State db f Sskip k e (Cminor.set_optvar optid vres le) m')

  | step_seq: forall db f s1 s2 k e le m,
      step (State db f (Sseq s1 s2) k e le m)
        E0 (State db f s1 (Kseq s2 k) e le m)

  | step_ifthenelse: forall db f a s1 s2 k e le m v b,
      eval_expr e le m a v ->
      Val.bool_of_val v b ->
      step (State db f (Sifthenelse a s1 s2) k e le m)
        E0 (State db f (if b then s1 else s2) k e le m)

  | step_loop: forall db f s k e le m,
      step (State db f (Sloop s) k e le m)
        E0 (State db f s (Kseq (Sloop s) k) e le m)

  | step_block: forall db f s k e le m,
      step (State db f (Sblock s) k e le m)
        E0 (State db f s (Kblock k) e le m)

  | step_exit_seq: forall db f n s k e le m,
      step (State db f (Sexit n) (Kseq s k) e le m)
        E0 (State db f (Sexit n) k e le m)
  | step_exit_block_0: forall db f k e le m,
      step (State db f (Sexit O) (Kblock k) e le m)
        E0 (State db f Sskip k e le m)
  | step_exit_block_S: forall db f n k e le m,
      step (State db f (Sexit (S n)) (Kblock k) e le m)
        E0 (State db f (Sexit n) k e le m)

  | step_switch: forall db f a cases k e le m n,
      eval_expr e le m a (Vint n) ->
      step (State db f (Sswitch a cases) k e le m)
        E0 (State db f (seq_of_lbl_stmt (select_switch n cases)) k e le m)

  | step_return_0: forall db f k e le m m',
      region_free_aux (Zsucc db) m (List.map snd f.(fn_vars)) = Some m' ->
      step (State db f (Sreturn None) k e le m)
        E0 (Returnstate Vundef (call_cont k) m')
  | step_return_1: forall db f a k e le m v m',
      eval_expr e le m a v ->
      region_free_aux (Zsucc db) m (List.map snd f.(fn_vars)) = Some m' ->
      step (State db f (Sreturn (Some a)) k e le m)
        E0 (Returnstate v (call_cont k) m')

  | step_label: forall db f lbl s k e le m,
      step (State db f (Slabel lbl s) k e le m)
        E0 (State db f s k e le m)

  | step_goto: forall db f lbl k e le m s' k',
      find_label lbl f.(fn_body) (call_cont k) = Some(s', k') ->
      step (State db f (Sgoto lbl) k e le m)
        E0 (State db f s' k' e le m)

  | step_internal_function: forall bf f vargs k m m1 e le,
      Genv.find_symbol ge f.(fn_id) = Some bf ->
      Genv.find_funct_ptr ge bf = Some (Internal f) ->
      list_norepet (map fst f.(fn_vars)) ->
      list_norepet f.(fn_params) ->
      list_disjoint f.(fn_params) f.(fn_temps) ->
      forall m0 b osz, forall DUMMYBLOCK:
               Mem.alloc m 0 0
                         (Mem.Tag_stack
                            (Mem.Stack_main f.(fn_id) osz)) = (m0, b),
               alloc_variables m0 b f.(fn_id) (fn_vars f) e m1 osz ->
               forall CALLCOND: call_condition f.(fn_vars),
               forall ALIGN: forall id sz al of, In (id, (sz, al, of)) (f.(fn_vars)) ->
                                                 (0 <= sz /\ 0 < al /\ (al | sz) /\ (al | 8)),
      bind_parameters f.(fn_params) vargs (create_undef_temps f.(fn_temps)) = Some le ->
      step (Callstate (Internal f) vargs k m)
        E0 (State b f f.(fn_body) k e le m1)

  | step_external_function: forall ef vargs k m t vres m',
      external_call ef ge vargs m t vres m' ->
      step (Callstate (External ef) vargs k m)
         t (Returnstate vres k m')        

  | step_return: forall db v optid f e le k m,
      step (Returnstate v (Kcall db optid f e le k) m)
        E0 (State db f Sskip k e (Cminor.set_optvar optid v le) m)
.

End RELSEM.

(** Execution of whole programs are described as sequences of transitions
  from an initial state to a final state.  An initial state is a [Callstate]
  corresponding to the invocation of the ``main'' function of the program
  without arguments and with an empty continuation. *)

Inductive initial_state (p: program): state -> Prop :=
  | initial_state_intro: forall b f m0,
      let ge := Genv.globalenv p in
      Genv.init_mem p = Some m0 ->
      Genv.find_symbol ge p.(prog_main) = Some b ->
      Genv.find_funct_ptr ge b = Some f ->
      funsig f = mksignature nil (Some Tint) ->
      initial_state p (Callstate f nil Kstop m0).

(** A final state is a [Returnstate] with an empty continuation. *)

Inductive final_state: state -> int -> Prop :=
  | final_state_intro: forall r m,
      final_state (Returnstate (Vint r) Kstop m) r.

(** Wrapping up these definitions in a small-step semantics. *)

Definition semantics (p: program) :=
  Semantics step (initial_state p) final_state (Genv.globalenv p).

End WITH_OFFSETS.

End WITHMEM.
